from .point_to_area import *
from . import point_to_area
from . import mode
from .mode import *
from .saller import sal
from . import vgm
from . vgm import *
from .rigider import *
from . import rigider
from .fss import *
from . import fss
from .significance import *
from . import significance
from . import acc
from .acc import *
from . import fuzzy_logic
from .fqi import fqi
from .uqi import uqi
from .geometric_characterizations.aindex import area_index
from .geometric_characterizations.cindex import count_index
from .geometric_characterizations.sindex import surround_index
from .hausdorff_metric.locperf import  *
from .pphindcast.pphindcast2d import Practically_Perfect_Hindcast
from .s1.S1 import *
from .geo_box_plot.GeoBoxPlot import Geo_Box_Plot
from . import cra


from meteva.base import IV
import numpy as np


def mre_skill(mre, mre_base):
    '''
    mre_skill 相对误差技巧，其中mre是精细化网格预报竞赛检验办法中的水量定量相对误差检验指标，
    :param mre:
    :param mre_base:
    :return: 相对技巧值，取值范围为负无穷到1，但负无穷记为-999返回
    '''
    if mre_base == 0:
        if mre > 0:
            return -999
        else:
            return 0
    else:
        smre1 = (mre_base - mre) / mre_base
        return smre1

def sme(mre, mre_base):
    '''
    :param mre: 省台的定量降水相对误差
    :param mre_base: 中央台的定量降水相对误差
    :return: 省台相对于中央台的技巧
    '''
    if mre_base == 0:
        if mre > 0:
            return -IV
        else:
            return 0
    else:
        skill = (mre_base - mre) / mre_base
        return round(skill,3)

def sst(mae,mae_base):
    '''

    :param mae: 省台的平均绝对误差
    :param mae_base: 中央台的平均绝对误差
    :return: 省台相对于中央台的技巧
    '''


    if mae_base == 0:
        skill = 1.01
    else:
        skill = (mae_base - mae)/(mae_base)

    return round(skill,3)


def tbask(ob,fo):

    '''
    根据输入的观测和预报数据计算预报员的基本评分
    ob: 观测数据序列
    fo: 预报数据序列，包含客观预报和预报员预报供两列
    '''
    fo_forecaster = fo[0, :]  # 提取预报员的预报
    error_forecaster = np.abs(fo_forecaster - ob)  # 计算绝对误差
    score_forecaster = np.zeros(error_forecaster.shape)  # 初始化一个基本评分数组
    index = np.where(error_forecaster <= 2)
    score_forecaster[index] = 0.6 # 将误差符合条件站次设为0.6分，其它继续为0

    # 提取模式的预报#############技巧评分
    fo_model = fo[1, :]
    error_model = np.abs(fo_model - ob)  # 计算绝对误差
    skill = np.zeros(ob.shape)  # 初始化一个基本评分数组

    index = np.where((error_model <= 2) & (error_forecaster <= 2))
    skill[index] = 0.6
    index = np.where((error_model >= 2) & (error_forecaster >= 2))
    skill[index] = 0.3
    index = np.where((error_model > 2) & (error_forecaster <= 2))
    skill[index] = 1

    score = score_forecaster + skill

    score_sum = np.sum(score)  # 计算总样本的平均分

    tbask = np.array([ob.size,score_sum])
    return tbask  # 返回总得分

def temp_forecaster_score(ob,fo):


    '''
    根据输入的观测和预报数据计算预报员的基本评分
    ob: 观测数据序列
    fo: 预报数据序列，包含客观预报和预报员预报供两列
    '''

    tbask_array = tbask(ob,fo)
    score = tbask_array[1]/tbask_array[0]
    return score  # 返回总得分


def temp_forecaster_score_tbask(tbask_array):


    score = tbask_array[1]/tbask_array[0]
    return score  # 返回总得分


import numpy as np
import meteva
from scipy.ndimage import convolve
from meteva.method.space.mode.distmap import getRedDotsCoordinatesFromLeftToRight
from meteva.method.space.mode.feature_axis import convexHull
from shapely.geometry import Polygon
import copy

def area_index(grd_ob,grd_fo,thresholds):
    '''

    :param grd_ob: 观测网格数据
    :param grd_fo: 预报网格数据
    :param threholds: 等级阈值
    :return:
    '''
    x = copy.deepcopy(grd_ob.values.squeeze())
    aindex_x = aindex(x,thresh=thresholds)
    y = copy.deepcopy(grd_fo.values.squeeze())
    aindex_y = aindex(y, thresh=thresholds)
    resulut = {"area_index":[aindex_x["area_index"],aindex_y["area_index"]],
               "total_area":[aindex_x["total_area"],aindex_y["total_area"]],
               "total_convex_area":[aindex_x["total_convex_area"],aindex_y["total_convex_area"]]}
    return resulut

def aindex(x, thresh=None,dx = 1,dy = 1):
    if thresh is None:
        thresh = 1e-08
    sx = np.zeros(x.shape)
    sx[x >= thresh] = 1
    kernel =    np.array([[1, 1, 1],
                         [1, 1, 1],
                        [1, 1, 1]])/9
    sm =  convolve(sx, kernel) #均值滤波
    #sm = sx
    thresh1 = 50/255   #平滑后亮度大于50
    sm1 = np.zeros(x.shape)
    sm1[sm >= thresh1] = 1
    area_all = np.count_nonzero(sx) * dx * dy
    label_array = meteva.method.mode.measure_label(sm1)  #转换成label矩阵
    num_obj =int(np.max(label_array))  #连通域的数量
    area_list = []
    for i in range(1,num_obj+1):
        label_one = np.zeros(x.shape)
        label_one[label_array==i] = 1
        rx = getRedDotsCoordinatesFromLeftToRight(label_one)   #所以label的格点坐标
        pts = convexHull(rx)     #求凸包络线
        polygon = Polygon(pts)          #计算包络线面积
        area_list.append(polygon.area)
    area_array = np.array(area_list)
    #a_convex = np.max(area_array)  # 获取包络线面积最大值
    a_convex = np.sum(area_array) # 获取包络线面积和
    area_index = area_all / a_convex   # 所有目标的格点面积和 与最大包络面积之比
    #res = pd.DataFrame({"area_index": area_index, "area_all": area_all, "Aconvex": a_convex, "dx": dx, "dy": dy}, index=[0])
    res = {"area_index": area_index, "total_area": area_all, "total_convex_area": a_convex}
    return res





# def aindex_cv(x, thresh=None, dx=1, dy=1):
#     if thresh is None:
#         thresh = 1e-08
#     sx = x
#     sx[sx >= thresh] = 255
#     sx[sx < thresh] = 0
#     cv2.imwrite("pic/aindex/data.png", sx)
#     img = cv2.imread("pic/aindex/data.png")
#     gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
#     blur = cv2.blur(gray, (3, 3))
#     ret, thresh = cv2.threshold(blur, 50, 255, cv2.THRESH_BINARY)
#     contours, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)  #
#     hull = []
#     for i in range(len(contours)):
#         hull.append(cv2.convexHull(contours[i], False))
#     drawing = np.zeros((thresh.shape[0], thresh.shape[1], 3), np.uint8)
#     for i in range(len(contours)):
#         color_contours = (0, 255, 0)  # green - color for contours
#         color = (255, 0, 0)  # blue - color for convex hull
#         cv2.drawContours(drawing, hull, i, color, 1, 8)
#     cv2.imwrite("pic/aindex/mask.png", drawing)
#     cv2.imwrite("pic/aindex/all.png", cv2.add(drawing, img))
#     a = np.count_nonzero(sx) * dx * dy
#     a_convex = cv2.contourArea(hull[0]) * dx * dy  # 面积最大的目标的包络线内的面积
#     a_index = a / a_convex
#     res = pd.DataFrame({"Aindex": a_index, "A": a, "Aconvex": a_convex, "dx": dx, "dy": dy}, index=[0])
#     return res


if __name__ == '__main__':
    #geom000 = pyreadr.read_r('./data/geom000.Rdata')['geom000']
    # sns.heatmap(pd.DataFrame(geom000))
    # plt.show()
    # sx = geom000.values
    # thresh = 1e-08
    # sx[sx >= thresh] = 255
    # sx[sx < thresh] = 0
    # image = Image.fromarray(sx)
    # image.show()

    import meteva.base as meb
    import math
    import meteva.method as mem
    # grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    # path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    # grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    # obs_array = grd_ob.values.squeeze()
    # res = aindex(obs_array)
    # print(res)

    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    dat_ob = np.zeros((grid1.nlat, grid1.nlon))
    for j in range(grid1.nlat):
        for i in range(grid1.nlon):
            dat_ob[j, i] = 20 * math.exp(-0.001 * (i - 200) ** 2 - 0.001 * (j - 200) ** 2)
    grd_ob = meb.grid_data(grid1, dat_ob)

    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    dat_fo = np.zeros((grid1.nlat, grid1.nlon))
    for j in range(grid1.nlat):
        for i in range(grid1.nlon):
            dat_fo[j, i] = 10 * math.exp(-0.001 * (i - 240) ** 2 - 0.0003 * (j - 240) ** 2)
            dat_fo[j, i] += 10 * math.exp(-0.001 * (i - 170) ** 2 - 0.0003 * (j - 170) ** 2)
    grd_fo = meb.grid_data(grid1, dat_fo)

    result = mem.space.area_index(grd_ob,grd_fo,thresholds=5)
    print(result)


import math
import numpy as np
import time


def frprmn2(x,targe,grads,tol = 0.001,caculate_time = 60,show = False):
    x0 = x
    n = len(x0)
    x_min = math.sqrt(1.0/n)
    df = grads(x0)
    g_0 = -df
    h_0 = g_0
    f_min_line0 = 0
    f_min_line1 = targe(x0)
    start_time = time.time()
    for i in range(10000):
        d0 = np.dot(df,h_0)
        x1 = x0 + h_0 * x_min
        df = grads(x1)
        d1 = np.dot(df,h_0)
        if d0 == d1:
            f_min_line1 = targe(x0)
            return x0
        else:
            x_min = x_min * d0/(d0 - d1)
            x0 += h_0 * x_min
        df = grads(x0)
        g_1 = -df
        v1 = np.dot(g_1 - g_0, g_1)
        v2 = np.dot(g_0,g_0)
        h_0 = g_1 + (v1/v2) * h_0
        g_0 = g_1
        if i%10 == 0:
            f_min_line0 = f_min_line1
            f_min_line1 = targe(x0)
            if show:print(str(i) + "  " + str(f_min_line1))
            delta = 2 * abs(f_min_line1 - f_min_line0) / (abs(f_min_line1) + abs(f_min_line0)) - tol
            if delta < 0:
                return x0
        if(time.time() - start_time > caculate_time):
            return x0
    print('迭代分析超过10000次而退出')
    return x0




from . import application
from . import program
from . import presentation
from . import regulation
from .program import *
from .application import *
from .presentation import *
from .regulation import *





# -*-coding:utf-8-*-

import math
import numpy as np


# def distmap(a):
#     xcol = a["xcol"]
#     yrow = a["yrow"]
#     m = np.array(a["m"])
#     rlen, clen = m.shape
#     true_list = np.array([])
#     dist = np.zeros((rlen, clen))
#     if sum(sum(m)) == 0:
#         distance = math.sqrt((xcol.max() - xcol.min())**2 + (yrow.max() - yrow.min()) ** 2)
#         dist = np.ones((rlen, clen)) * distance
#     elif sum(sum(m)) != m.size:
#         for i in range(rlen):
#             for j in range(clen):
#                 if m[i, j]:
#                     true_list = np.append(true_list, [xcol[i], yrow[j]])
#         for i in range(rlen):
#             for j in range(clen):
#                 if m[i, j]:
#                     dist[i, j] = 0
#                 else:
#                     point = np.array([xcol[i], yrow[j]])
#                     mat = spatial.distance.cdist(point, true_list, 'cityblock')
#                     dist[i, j] = mat.min()
#     return dist


def rebound(a, window):
    w = a
    x_range = window["x_range"]
    y_range = window["y_range"]
    xcol = newseq(a["xcol"], x_range)
    yrow = newseq(a["yrow"], y_range)
    m = np.zeros((len(xcol), len(yrow)))
    x_min_index = list(xcol).index(min(a['xcol']))
    x_max_index = list(xcol).index(max(a['xcol']))
    y_min_index = list(yrow).index(min(a['yrow']))
    y_max_index = list(yrow).index(max(a['yrow']))
    m[x_min_index:x_max_index+1, y_min_index:y_max_index+1] = a['m']
    m = (m == 1)
    newmask = {"xcol": xcol, "yrow": yrow, "m": m}
    # xx = rasterx(newmask)
    # yy = rastery(newmask)
    # newmask["m"] = inside(xx, yy, w)
    return newmask


## 用于mask
def boundingbox(a, b):
    a_x_min = a["xrange"][0]
    a_x_max = a["xrange"][1]
    a_y_min = a["yrange"][0]
    a_y_max = a["yrange"][1]
    b_x_min = b["xrange"][0]
    b_x_max = b["xrange"][1]
    b_y_min = b["yrange"][0]
    b_y_max = b["yrange"][1]
    x_min = min(a_x_min, b_x_min)
    x_max = max(a_x_max, b_x_max)
    y_min = min(a_y_min, b_y_min)
    y_max = max(a_y_max, b_y_max)
    x_range = [x_min, x_max]
    y_range = [y_min, y_max]
    window = {"x_range": x_range, "y_range": y_range}
    return window


def newseq(oldseq, newrange):
    oldrange = [min(oldseq), max(oldseq)]
    length = len(oldseq)
    dstep = (max(oldseq) - min(oldseq)) / (length - 1)
    nleft = max(0, math.floor((oldrange[0] - newrange[0]) / dstep))
    nright = max(0, math.floor((newrange[1] - oldrange[1]) / dstep))
    newstart = max(oldrange[0] - nleft * dstep, newrange[0])
    newend = min(oldrange[1] + nright * dstep, newrange[1]) + dstep
    return np.arange(newstart, newend, dstep)


# def rasterx(w):
#     len_y = w["yrow"].size
#     return np.tile(np.array(w["xcol"]), (len_y, 1))


# def rastery(w):
#     len_x = w["xcol"].size
#     return np.tile(np.array(w["yrow"]), (len_x, 1)).transpose()


# def inside(x, y, w):
#     x = np.array(x)
#     y = np.array(y)
#     xr = w["xrange"]
#     yr = w["yrange"]
#     frameok = np.array((x >= xr[0]) * (x <= xr[1]) * (y >= yr[0]) * (y <= yr[1]))
#     xf = x.transpose()[frameok.transpose()]
#     yf = y.transpose()[frameok.transpose()]
#     loc = nearest_raster_point(xf, yf, w)
#     okf = np.zeros(loc['col'].size)
#     for i in range(loc['col'].size):
#         okf[i] = w['m'][int(loc['row'][i]) - 1, int(loc['col'][i]) - 1]
#     result = frameok[frameok]
#     for i in range(okf.size):
#         result[i] = okf[i]
#     return okf.reshape(w['m'].shape)


# def nearest_raster_point(x, y, w):
#     x = np.array(x)
#     y = np.array(y)
#     nr = w["dim"][0]
#     nc = w["dim"][1]
#     cc = 1 + np.round((x - w["xcol"][0]) / w["xstep"])
#     rr = 1 + np.round((y - w["ycol"][0]) / w["ystep"])
#     cc = np.maximum(1, np.minimum(cc, nc))
#     rr = np.maximum(1, np.minimum(rr, nr))
#     return {"row": rr, "col": cc}


def rotate(self, angle):
    x = math.cos(angle) * self[:, 0] - math.sin(angle) * self[:, 1]
    y = math.sin(angle) * self[:, 0] + math.cos(angle) * self[:, 1]
    # result = {'xrange': [np.min(x), np.max(x)], 'yrange': [np.min(y), np.max(y)], 'x0': np.min(x), 'x1': np.max(x), 'y0': np.min(y), 'y1': np.max(y)}
    result = {'xrange': [np.min(x), np.max(x)], 'yrange': [np.min(y), np.max(y)],
              'x0': x[0], 'x1': x[1], 'y0': y[0], 'y1': y[1]}
    return result


def getRedDotsCoordinatesFromLeftToRight(np_matrix, red_dor_number=1):
    red_dots = np.where(np_matrix == red_dor_number)
    red_dots = tuple(g[np.argsort(red_dots[-1])] for g in red_dots)
    red_dots = np.stack(red_dots)[::-1].T
    return red_dots



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 24 16:39:08 2022

"""
import numpy as np
from scipy import stats
import sys

#sys.path.append(r'/Users/tangbuxing/Desktop/tra_test-SAL/fieldSignficance/lib')
from meteva.method.space.field_significance import sig_coverage
#import sig_coverage


def sig_cor_t(r, Len=40):
    t = abs(r) * np.sqrt((Len - 2) / (1 - r ** 2))
    # qt -> stats.t(df=359).ppf((0.84473448, 0.15382939))
    palpha_cor = 1 - stats.t.cdf(t, df=Len - 2)  # t分布的分布函数
    return palpha_cor


def MCdof(x, ntrials=1000, field_sig=0.05, zfun="rnorm", zfun_args='NULL',
          which_test=["t", "Z", "cor.test"], show=False):

    output = {}
    if len(which_test) > 1:
        which_test = "t"
    x = np.array(x)
    xdim = x.shape
    tlen = xdim[0]
    B_dof_test = np.zeros(ntrials)
    if show:
        print("\n", "Looping through ", ntrials, " times to simulate data and take correlations.  Enjoy!\n")
    for i in range(1, ntrials):
        if show and (i < 100 or i % 100 == 0):
            print(i, " ")
        # rnorm(n, mean = 0, sd = 1)返回值是n个正态分布随机数构成的向量。
        mu, sigma = 0, 1  # mean and standard deviation
        z = np.random.normal(mu, sigma, tlen)  # 正态分布存在差异
        '''
        #绘制正态分布图
        count, bins, ignored = plt.hist(s, 30, density=True)
        plt.plot(bins, 1/(sigma * np.sqrt(2 * np.pi)) * np.exp( - (bins - mu)**2 / (2 * sigma**2) ),
                 linewidth=2, color='r')
        plt.show()
        '''
        # 判断为否，不执行
        tmp = np.array([])
        if which_test == "cor.test":
            # tmp = apply(x, 2, cortester, y = z)
            tmp = np.corrcoef(x)[1, 0]
        else:
            #默认进行t检验
            # 计算pearson相关系数
            cor = np.array([])
            for j in range(x.shape[1]):  # 循环变量有问题
                x_j = x[:, j]
                xz = np.array([x_j, z])
                cor_0 = np.corrcoef(xz)[1, 0]  # 目前存在偏差是由于z的结果不一致导致，方法本身经过验证是没问题的
                cor = np.append(cor, cor_0)
            cor_value = np.abs(cor)
            # 只翻译默认的第一个参数
            if which_test == "t":
                tmp = sig_cor_t(r=cor_value, Len=tlen)  # 输出结果
            '''
            else if (which.test == "Z") 
                tmp <- sig.cor.Z(cor.value, len = tlen, ...)
            '''
        tmp_small = tmp[field_sig>tmp] #提取显著性超阈值那部分显著性

        # if tmp_small.size == 0:
        #     B_dof_test[i] = 1
        # else:
        #     B_dof_test[i] = np.mean(tmp_small)  # 显著性超阈值部分的均值
        B_dof_test[i] = tmp_small.size / tmp.size

    minsigcov_val = np.quantile(a=B_dof_test, q=1 - field_sig)
    output = {'MCprops': B_dof_test, 'minsigcov': minsigcov_val}
    return output


def is_sig(X, blockboot_result_df, n=1000, fld_sig=0.05, show=False):
    output = {}
    '''
    X = errfield
    blockboot_result_df = hold
    n = ntrials
    fld_sig = field_sig
    '''

    # sig_result 表示利用蒙特卡罗方法测试所有格点和随机数求相关后，表现为具有显著性的格点数占比。
    sig_result = MCdof(X, ntrials=n, field_sig=0.05, show=show)['minsigcov']

    #actual_coverage表示平均值具有显著性的格点数占比。
    actual_coverage = sig_coverage.sig_coverage(DF=blockboot_result_df)
    sig = actual_coverage > sig_result  #输出结果
    '''
    output <- list(name = as.character(deparse(substitute(X))), 
                   required = as.numeric(sig.results), actual = as.numeric(actual.coverage), 
                   issig = as.logical(sig))
    '''
    output = {"name": X, "required": sig_result, "actual": actual_coverage, "issig": sig}

    return output









import math
import meteva
import numpy as np
import copy
import pandas as pd
import datetime


def temp_decrease_in_process(sta,used_coords = "dtime"):
    if used_coords == "dtime":
        change_24 = change(sta,delta=24,used_coords="dtime")
        change_48 = change(sta,delta=48,used_coords="dtime")
        change_24_48 = meteva.base.min_on_level_time_dtime_id(change_24,change_48,how="outer",default=0)
        change_most_dtime =  meteva.base.loc_of_min(change_24_48,used_coords=["dtime"],ignore_missing=True)
        change_most_dtime.loc[:,"dtime"] = change_most_dtime.iloc[:,6]
        change_most = meteva.base.min_of_sta(change_24_48,used_coords=["dtime"],ignore_missing=True)
        change_most.attrs = copy.deepcopy(sta.attrs)
        return change_most

def accumulate_time(sta_ob,step,keep_all = True):
    '''
    观测数据累加
    :param sta_ob:
    :param step:
    :param keep_all:
    :return:
    '''

    times= sta_ob.loc[:,'time'].values
    times = list(set(times))
    times.sort()
    times = np.array(times)
    dtimes = times[1:] - times[0:-1]
    min_dtime = np.min(dtimes)
    rain_ac = None
    for i in range(step):
        rain1 = sta_ob.copy()
        rain1["time"] = rain1["time"] + min_dtime * i
        rain_ac = meteva.base.add_on_level_time_dtime_id(rain_ac,rain1,how="inner")
    if not keep_all:
        dtimes = times[:] - times[-1]
        dh = (dtimes/min_dtime).astype(np.int32)
        new_times = times[dh%step ==0]
        rain_ac = meteva.base.in_time_list(rain_ac,new_times)
    print("warning: accumulate_time函数将在后续升级中不再支持，请重新使用sum_of_sta函数满足相关需求")
    return rain_ac

def accumulate_dtime(sta,step,keep_all = True):
    '''观测数据累加'''

    dtimes= sta.loc[:,'dtime'].values
    dtimes = list(set(dtimes))
    dtimes.sort()
    dtimes = np.array(dtimes)
    dhour_unit = dtimes[0]
    if dhour_unit ==0:
        dhour_unit = dtimes[1]
    rain_ac = None
    for i in range(step):
        rain1 = sta.copy()
        rain1["dtime"] = rain1["dtime"] + dhour_unit * i
        #print(dhour_unit * i)
        rain_ac = meteva.base.add_on_level_time_dtime_id(rain_ac,rain1,how="inner")
    if not keep_all:
        dh =((dtimes - dtimes[-1])/dhour_unit).astype(np.int32)
        new_dtimes = dtimes[dh%step ==0]
        rain_ac = meteva.base.in_dtime_list(rain_ac,new_dtimes)
    return rain_ac

def change(data,delta = 24,used_coords = "time"):

    if used_coords == "time":
        if isinstance(data,pd.DataFrame):
            names_0 = meteva.base.get_stadata_names(data)
            names_1 = []
            for name in names_0:
                names_1.append(name + "_new")
            sta1 = data.copy()
            meteva.base.set_stadata_names(sta1, names_1)
            sta1["time"] = sta1["time"] + datetime.timedelta(hours= delta)
            sta01 = meteva.base.combine_on_all_coords(sta1, data)
            fn = len(names_1)
            dvalue = sta01.iloc[:, (-fn):].values - sta01.iloc[:, (-fn * 2):(-fn)].values
            sta01.iloc[:, (-fn):] = dvalue
            sta01 = sta01.drop(names_1, axis=1)
            sta01.attrs = copy.deepcopy(data.attrs)
            sta01.attrs["valid_time"] = delta
            return sta01
        else:
            grid0 = meteva.base.get_grid_of_data(data)
            times_all = data["time"].values
            index_dtime = {}
            time1_all = []
            for i in range(len(times_all)):
                time1 = meteva.base.all_type_time_to_datetime(times_all[i])
                time1_all.append(time1)
                index_dtime[time1] = i
            index_s = []
            index_e = []
            for i in range(len(times_all)):
                time_e = time1_all[i]
                time_s = time_e - datetime.timedelta(hours=delta)
                if time_s in index_dtime.keys():
                    index_e.append(i)
                    index_s.append(index_dtime[time_s])
            time_list_e = times_all[index_e]
            if len(time_list_e) ==0:
                gtime1 = time_list_e
            else:
                gtime1 = [time_list_e[0],time_list_e[-1],time_list_e[1] - time_list_e[0]]
            dat_delta = data.values[:,:,index_e,:,:,:] - data.values[:,:,index_s,:,:,:]
            grid1 = meteva.base.grid(grid0.glon,grid0.glat,gtime1,grid0.dtimes,grid0.levels,grid0.members)
            grd_change = meteva.base.grid_data(grid1,dat_delta)
            return grd_change
    else:
        if isinstance(data, pd.DataFrame):
            names_0 = meteva.base.get_stadata_names(data)
            names_1 = []
            for name in names_0:
                names_1.append(str(name)+"_new")
            sta1 = data.copy()
            meteva.base.set_stadata_names(sta1,names_1)
            sta1.loc[:,"dtime"] = sta1.loc[:,"dtime"] + delta
            sta01 = meteva.base.combine_on_all_coords(sta1,data)
            fn= len(names_1)
            dvalue = sta01.iloc[:,(-fn):].values - sta01.iloc[:,(-fn * 2):(-fn)].values
            sta01.iloc[:,(-fn):] = dvalue
            sta01 = sta01.drop(names_1,axis=1)
            sta01.attrs = copy.deepcopy(data.attrs)
            sta01.attrs["valid_time"] = delta
            return sta01
        else:
            grid0 = meteva.base.get_grid_of_data(data)
            dtimes_all = data["dtime"].values
            index_dtime = {}
            for i in range(len(dtimes_all)):
                index_dtime[dtimes_all[i]] = i
            index_s = []
            index_e = []
            for i in range(len(dtimes_all)):
                dtime_e = dtimes_all[i]
                dtime_s = dtime_e - delta
                if dtime_s in index_dtime.keys():
                    index_e.append(i)
                    index_s.append(index_dtime[dtime_s])
            dtime_list_e = dtimes_all[index_e]
            dat_delta = data.values[:,:,:,index_e,:,:] - data.values[:,:,:,index_s,:,:]
            grid1 = meteva.base.grid(grid0.glon,grid0.glat,grid0.gtime,dtime_list_e,grid0.levels,grid0.members)
            grd_change = meteva.base.grid_data(grid1,dat_delta)
            return grd_change

def t_rh_to_tw(temp,rh,rh_unit = "%"):
    '''根据温度和相对湿度计算湿球温度'''
    if isinstance(temp,pd.DataFrame):
        sta1 = meteva.base.combine_on_all_coords(temp, rh)
        meteva.base.set_stadata_names(sta1, ["t", "rh"])
        sta2 = meteva.base.not_IV(sta1)
        T = sta2.loc[:,"t"].values
        RH = sta2["rh"].values
        if(T[0]>120):
            T -= 273.16

        if rh_unit == "%":
            pass
        else:
            RH = RH * 100
        max_rh = np.max(RH)
        min_rh = np.min(RH)
        if max_rh>100 or min_rh <0:
            print("相对湿度取值不能超过100%或小于0%")
            return
        if max_rh < 1:
            print("警告：最大的相对湿度小于1%，请确认rh的单位是否为%，如果不是,请设置rh_unit = 1")

        Tw = T * np.arctan(0.151977 * np.sqrt(RH + 8.313659)) + np.arctan(T + RH) - np.arctan(
            RH - 1.676331) + 0.00391838 * np.power(RH, 1.5) * np.arctan(0.023101 * RH) - 4.686035

        sta2["tw"] = Tw
        sta = sta2.drop(["t", "rh"], axis=1)
        sta.attrs = copy.deepcopy(sta.attrs)
        sta.attrs["var_name"] = "tw"
        sta.attrs["var_cn_name"] = "湿球温度"
        sta.attrs["var_units"] = "degC"
        return sta
    else:
        grid0 = meteva.base.get_grid_of_data(temp)
        if temp.values[0,0,0,0,0,0] >120:
            T = temp.values - 273.16
        else:
            T = temp.values

        RH = rh.values
        if rh_unit == "%":
            RH /= 100
        else:
            pass
        max_rh = np.max(RH)
        min_rh = np.min(RH)
        if max_rh>1 or min_rh <0:
            print("相对湿度取值不能超过100%或小于0%")
            return
        if max_rh < 0.01:
            print("警告：最大的相对湿度小于1%，请确认rh的单位是否为%，如果不是,请设置rh_unit = 1")

        Tw = T * np.arctan(0.151977 * np.sqrt(RH + 8.313659)) + np.arctan(T + RH) - np.arctan(
            RH - 1.676331) + 0.00391838 * np.power(RH, 1.5) * np.arctan(0.023101 * RH) - 4.686035

        grd = meteva.base.grid_data(grid0,Tw)
        return grd

def u_v_to_speed_angle(u,v):
    '''
    将u，v 转换成风速，风向
    :param u:
    :param v:
    :return:
    '''
    if isinstance(u, pd.DataFrame):
        sta = meteva.base.combine_on_all_coords(u, v)
        datanames = meteva.base.get_stadata_names(sta)
        nu = int(len(datanames)/2)
        #nsta = len(sta.indexs)
        ud = sta.iloc[:,6:(6+nu)].values.astype(np.float32)
        vd = sta.iloc[:,(6+nu):].values.astype(np.float32)

        s,a = meteva.base.tool.math_tools.u_v_to_s_d(ud,vd)
        speed = sta.iloc[:,0:(6+nu)].copy()
        angle = speed.copy()
        speed.iloc[:,6:(6+nu)] = s[...]
        angle.iloc[:, 6:(6 + nu)] = a[...]
        names1 = []
        names2 = []
        for i in range(nu):
            names1.append("speed"+str(i))
            names2.append("angle"+str(i))
        meteva.base.set_stadata_names(speed,names1)
        meteva.base.set_stadata_names(angle,names2)
        speed.attrs = copy.deepcopy(u.attrs)
        speed.attrs["var_name"] = "wsp"
        sta.attrs["var_cn_name"] = "风速"
        sta.attrs["var_units"] = "m/s"
        angle.attrs = copy.deepcopy(u.attrs)
        angle.attrs["var_name"] = "wdir"
        angle.attrs["var_cn_name"] = "风向"
        angle.attrs["var_units"] = "degree"
        return speed,angle
    else:
        ud = u.values
        vd = v.values
        s, a = meteva.base.tool.math_tools.u_v_to_s_d(ud, vd)
        grid = meteva.base.get_grid_of_data(u)
        speed = meteva.base.grid_data(grid,s)
        angle = meteva.base.grid_data(grid,a)
        return speed,angle

def u_v_to_wind(u,v):
    if isinstance(u,pd.DataFrame):
        sta = meteva.base.combine_on_all_coords(u, v)
        meteva.base.set_stadata_names(sta, ["u", "v"])
        return  sta

    else:
        grid0 = meteva.base.get_grid_of_data(u)
        grid1 = meteva.base.grid(grid0.glon,grid0.glat,grid0.gtime,
                                                  dtime_list= grid0.dtimes,level_list=grid0.levels,member_list=["u","v"])
        wind = meteva.base.grid_data(grid1)
        wind.name = "wind"
        wind.values[0, :, :, :, :, :] = u.values[0, :, :, :, :, :]
        wind.values[1, :, :, :, :, :] = v.values[0, :, :, :, :, :]
        return wind

def wind_to_speed_angle(wind):
    if isinstance(wind,pd.DataFrame):
        member_name = meteva.base.get_stadata_names(wind)
        u = meteva.base.sele_by_para(wind, member=member_name[0::2])
        v =meteva.base.sele_by_para(wind, member=member_name[1::2])
        speed, angle = meteva.base.u_v_to_speed_angle(u, v)
        names = meteva.base.get_stadata_names(u)
        new_names = []
        for name in names:
            if name[0:2] == "u_":
                name = name[2:]
            new_names.append(name)
        meteva.base.set_stadata_names(speed,new_names)
        meteva.base.set_stadata_names(angle,new_names)
        return speed, angle
    else:
        u = meteva.base.in_member_list(wind,member_list=[0],name_or_index="index")
        v = meteva.base.in_member_list(wind, member_list=[1], name_or_index="index")
        speed,angle = u_v_to_speed_angle(u,v)
        return speed,angle

def speed_angle_to_wind(speed,angle = None):
    if isinstance(speed, pd.DataFrame):
        if angle is not None:
            sta = meteva.base.combine_on_all_coords(speed, angle)
        else:
            sta = speed.copy()
        meteva.base.set_stadata_names(sta, ["speed", "angle"])
        #speed = sta["speed"].values.astype(np.float32)
        #angle = sta["angle"].values.astype(np.float32)
        speed = sta["speed"].values.astype(np.float32)
        angle = sta["angle"].values.astype(np.float32)
        u = -speed * np.sin(angle  * 3.14 / 180)
        v = -speed * np.cos(angle * 3.14 / 180)
        sta["u"] = u
        sta["v"] = v
        sta = sta.drop(["speed", "angle"], axis=1)
        return sta



    else:
        speed_v = speed.values.squeeze()
        angle_v = angle.values.squeeze()
        grid0 = meteva.base.get_grid_of_data(speed)
        grid1 = meteva.base.grid(grid0.glon,grid0.glat,grid0.gtime,
                                                  dtime_list=grid0.dtimes,level_list=grid0.levels,member_list=["u","v"])
        wind = meteva.base.grid_data(grid1)
        wind.name = "wind"
        wind.values[0, :, :, :, :, :] = speed_v[:, :] * np.cos(angle_v[:, :] * math.pi /180)
        wind.values[1, :, :, :, :, :] = speed_v[:, :] * np.sin(angle_v[:, :] * math.pi /180)
        return wind

def t_dtp_to_rh(temp,dtp):
    if isinstance(temp,pd.DataFrame):
        sta = meteva.base.combine_on_all_coords(temp, dtp)
        meteva.base.set_stadata_names(sta, ["t", "dtp"])
        T = sta.loc[:,"t"].values
        if(T[0]>120):
            T -= 273.16

        D = sta["dtp"].values
        if D[0] >120:
            D -= 273.16
        e0 = 6.11 * np.exp(17.15 * T/(235 + T))
        e1 = 6.11 * np.exp(17.15 * D / (235 + D))

        rh = 100 * e1/e0
        sta["rh"] = rh
        sta = sta.drop(["t", "dtp"], axis=1)
        return sta
    else:
        grid0 = meteva.base.get_grid_of_data(temp)

        if temp.values[0,0,0,0,0,0] >120:
            T = temp.values - 273.16
        else:
            T = temp.values
        if dtp.values[0,0,0,0,0,0] >120:
            D = dtp.values - 273.16
        else:
            D = dtp.values

        e0 = 6.11 * np.exp(17.15 * T/(235 + T))
        e1 = 6.11 * np.exp(17.15 * D / (235 + D))

        rh = e1/e0
        grd = meteva.base.grid_data(grid0,rh)

        return grd

def t_rh_p_to_q(temp,rh,pressure,rh_unit = "%",check = False):
    '''
    根据温度、相对湿度和气压计算比湿
    :param temp: 温度，可以是摄氏度，也可以是绝对温度
    :param rh:  相对湿度，可以是0-100，也可以是0-1
    :param level: 气压，单位百帕,可以是整数，站点数据或网格数据
    :return:
    '''
    if isinstance(temp,pd.DataFrame):
        if not isinstance(pressure,pd.DataFrame):
            level_s = temp.copy()
            level_s.iloc[:,-1] = pressure
        else:
            level_s = pressure
        sta1 = meteva.base.combine_on_all_coords(temp, rh)
        sta2 = meteva.base.combine_on_all_coords(sta1, level_s)
        meteva.base.set_stadata_names(sta2, ["t", "rh","p"])
        sta2 = meteva.base.not_IV(sta2)
        T = sta2.loc[:,"t"].values
        R = sta2.loc[:,"rh"].values
        P = sta2.loc[:,"p"].values
        if(T[0]>120):
            T -= 273.16
        e0 = 6.11 * np.exp(5420 * (1.0 / 273.15 - 1 / (T + 273.15))) * 622

        if rh_unit == "%":
            R = R.astype(np.float)
            R = R/100
        else:
            pass

        if check:
            max_rh = np.max(R)
            min_rh = np.min(R)
            if max_rh>1 or min_rh <0:
                print("相对湿度取值不能超过100%或小于0%")
                return
            if max_rh < 0.01:
                print("警告：最大的相对湿度小于1%，请确认rh的单位是否为%，如果不是,请设置rh_unit = 1")

        q = e0 * R/P
        sta2["q"] = q
        sta = sta2.drop(["t", "rh","p"], axis=1)
        return sta
    else:
        grid0 = meteva.base.get_grid_of_data(temp)
        if temp.values[0,0,0,0,0,0] >120:
            T = temp.values - 273.16
        else:
            T = temp.values


        R = rh.values
        if rh_unit == "%":
            R /= 100
        else:
            pass
        if check:
            max_rh = np.max(R)
            min_rh = np.min(R)
            if max_rh>2 or min_rh <0:
                print("相对湿度最大值为"+str(max_rh))
                print("相对湿度最小值为" + str(min_rh))
                print("相对湿度取值不能超过100%或小于0%")
                return None

        e0 = 6.11 * np.exp(5420 * (1.0 / 273.15 - 1 / (T + 273.15))) * 622

        if isinstance(pressure,float) or isinstance(pressure,int):
            P = pressure
        else:
            P = pressure.values
        q = e0 * R/P
        grd = meteva.base.grid_data(grid0,q)
        return grd


def t_rh_to_vp(temp,rh):
    if isinstance(temp,pd.DataFrame):
        sta = meteva.base.combine_on_all_coords(temp, rh)
        meteva.base.set_stadata_names(sta, ["t", "rh"])
        T = sta["t"].values
        R = sta["rh"].values
        if(T[0]>120):
            T -= 273.16
        #e0 = 6.11 * np.exp(17.15 * T/(235 + T))
        e0 = 6.11 * np.exp(5420 * (1.0 / 273.15 - 1 / (T + 273.15))) * 622
        max_rh = np.max(R)
        if max_rh >1.1:
            R /= 100
        vp = e0 * R
        sta["vp"] = vp
        sta = sta.drop(["t", "rh"], axis=1)
        return sta
    else:
        grid0 = meteva.base.get_grid_of_data(temp)
        if temp.values[0,0,0,0,0,0] >120:
            T = temp.values - 273.16
        else:
            T = temp.values
        max_rh = np.max(rh.values)
        if max_rh >1.1:
            R = rh.values /100
        else:
            R = rh.values
        #e0 = 6.11 * np.exp(17.15 * T / (235 + T))
        e0 = 6.11 * np.exp(5420 * (1.0 / 273.15 - 1 / (T + 273.15))) * 622
        vp = e0 * R
        grd = meteva.base.grid_data(grid0,vp)
        return grd


def t_q_p_to_rh(temp,q,pressure,unit = "%"):
    '''
    根据温度、比湿和气压计算相对湿度
    :param temp:
    :param q:
    :param pressure:
    :return:
    '''
    rh100 = copy.deepcopy(temp)
    if isinstance(temp, pd.DataFrame):
        rh100.iloc[:,-1] = 100
    else:
        rh100.values[...] = 100
    q100 = t_rh_p_to_q(temp,rh100,pressure)
    if isinstance(temp, pd.DataFrame):
        rh = meteva.base.divide_on_level_time_dtime_id(q,q100)
        if unit == "%":
            rh.iloc[:,6:] *= 100
    else:
        rh = copy.deepcopy(temp)
        rh.values = q.values/q100.values
        #rh.values[rh.values>1] = 1
        if unit=="%":
            rh.values *= 100
    return rh





# 采用以下形式将所有.py 模块导入到mode 目录下
#from . import centdist
#from . import data_pre
#from . import imomenter
from . import saller

# 采用如下形式将5个用户必要的函数导入到mode模块下
#from .centdist import *
#from .imomenter import *
#from .data_pre import *
from .saller import sal




from . import score
from . import table
from . import plot
from . import skill

from .table import *
from .plot import scatter_regress,pdf_plot,box_plot_continue,taylor_diagram,frequency_histogram_error,accumulation_change_with_strenght,frequency_change_with_strenght
from .skill import mre_skill,sme,sst
from .score import sample_count,ob_mean,fo_mean,tase,me,mae,mse,rmse,me_tase,mae_tase,mse_tase,rmse_tase,tc_count
from .score import bias_m,bias_tmmsss,corr,corr_tmmsss,tmmsss,tmmsss_merge,toar,nse,nse_tase_tmmsss,nse_tmmsss,mre,mre_toar,correct_rate,correct_rate_tc
from .score import ob_fo_mean,ob_fo_std,ob_fo_max,ob_fo_min,ob_fo_cv
from .score import wrong_rate,wrong_rate_tc,residual_error,residual_error_tmmsss,residual_error_rate,residual_error_rate_tmmsss
from .score import max_abs_error,max_error,min_error
from .score import ob_fo_sum,rmsf,tlfo,corr_rank,rmsf_tlfo,ob_fo_quantile
from .score import ob_fo_sum_tmmsss,ob_fo_mean_tmmsss,ob_fo_std_tmmsss
from .score import pas,pmse
from .score import ob_fo_precipitation_strength,ob_fo_precipitation_strength_cscs,cscs,ob_fo_precipitation_strenght


from meteva.method.space.baddeley_binary_image_metric.locmeasures2d_prep import locmeasures2d_prep
from meteva.method.space.baddeley_binary_image_metric.lib.datagrabber_spatialVx import datagrabber_spatialVx
from meteva.method.space.baddeley_binary_image_metric.lib.deltametric import deltametric
from meteva.method.space.baddeley_binary_image_metric.lib.im import im
from meteva.method.space.baddeley_binary_image_metric.lib.loc_list_setup import loc_list_setup
from meteva.method.space.baddeley_binary_image_metric.lib.locperf import locperf
from meteva.method.space.baddeley_binary_image_metric.lib.solutionset import solutionset
import numpy as npy
import copy
#import pyreadr
from meteva.method.space.baddeley_binary_image_metric.lib.make_spatialVx import make_spatialVx


def locmeasures2d_spatial_vx(input_object, which_stats=None,
                             k=None, alpha=None, bdconst=None,
                             p=None, time_point=1, obs=1, model=1):
    if p is None:
        p = [2]
    if alpha is None:
        alpha = [0.1]
    if which_stats is None:
        which_stats = ["bdelta", "haus", "qdmapdiff", "med", "msd", "ph", "fom"]
    distfun = "distmapfun"
    distfun_params = None
    input_object = locmeasures2d_prep(input_object=input_object, k=k, alpha=alpha, bdconst=bdconst, p=p)
    a = input_object
    thresholds = a["thresholds"]
    q = thresholds[0].shape[0]
    if a["k"] is None and "qdmapdiff" in which_stats:
        print("locmeasures2d: must supply k in call to locmeasures2dPrep to do qdmapdiff method.")
        return
    try:
        nk = len(a["k"])
    except Exception:
        nk = 1
    try:
        np = len(a["p"])
    except Exception:
        np = 1
    try:
        nalpha = 1
    except Exception:
        nalpha = len(a["alpha"])
    out = loc_list_setup(a=a, which_stats=which_stats, nthresh=q, n_p=np, nk=nk, nalpha=nalpha)
    out["time_point"] = time_point
    out["model"] = model - 1
    dat = datagrabber_spatialVx(input_object, time_point=time_point, obs=obs - 1, model=model - 1)
    Obs = dat["X"]
    Fcst = dat["Xhat"]
    mainname = a["dataname"]
    vxname = a["obsname"][obs - 1]
    out["dataname"] = [mainname, vxname, a['modelname']]
    out["thresholds"] = {'X': thresholds[0][:, obs - 1], 'Xhat': thresholds[1][:, model - 1]}
    xdim = a["xdim"]
    x_id = im(Obs)
    y_id = im(Fcst)
    for threshold in range(0, q):
        x_res = copy.deepcopy(x_id)
        y_res = copy.deepcopy(y_id)
        x_res['v'] = x_id['v'] >= thresholds[0][threshold, obs - 1]
        y_res['v'] = y_id['v'] >= thresholds[1][threshold, model - 1]
        Ix = solutionset(x_res)
        Iy = solutionset(y_res)
        if "bdelta" in which_stats:
            for p in range(0, np):
                tmpDelta = deltametric(Iy, Ix, p=a['p'][p], c=a['bdconst'])
                if type(tmpDelta) != "try-error":
                    out['bdelta'][p, threshold] = tmpDelta
        if "haus" in which_stats:
            out['haus'][threshold] = deltametric(Iy, Ix, p=float("inf"), c=float("inf"))
        if "qdmapdiff" in which_stats:
            for k in range(0, nk):
                out['qdmapdiff'][k, threshold] = \
                    locperf(X=Ix, Y=Iy, which_stats="qdmapdiff", k=a['k'][k], distfun=distfun,
                            distfun_params=distfun_params)['qdmapdiff']
        w_id = ("med" in which_stats or "msd" in which_stats)
        if w_id:
            tmp = locperf(X=Ix, Y=Iy, which_stats=["med", "msd"], distfun=distfun, distfun_params=distfun_params)
            if "med" in which_stats:
                out['medMiss'][threshold] = tmp['medMiss']
                out['medFalseAlarm'][threshold] = tmp['medFalseAlarm']

            if "msd" in which_stats:
                out['msdMiss'][threshold] = tmp['msdMiss']
                out['msdFalseAlarm'][threshold] = tmp['msdFalseAlarm']

        if "fom" in which_stats:
            for i in range(1, nalpha):
                out['fom'][i, threshold] = locperf(X=Ix, Y=Iy, which_stats="fom", alpha=a['alpha'][i], distfun=distfun,
                                                   distfun_params=distfun_params)['fom']

    return out


if __name__ == '__main__':
    # geom000 = pyreadr.read_r('./data/geom000.Rdata')['geom000']
    # geom001 = pyreadr.read_r('./data/geom001.Rdata')['geom001']
    # ICPg240Locs = pyreadr.read_r('./data/ICPg240Locs.Rdata')['ICPg240Locs']

    import meteva.base as meb
    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    path_fo_03 = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    path_fo_27 = r'H:\test_data\input\mem\mode\ec\rain03\20072508.027.nc'
    grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    grd_fo_03 = meb.read_griddata_from_nc(path_fo_03, grid=grid1, time="2020072608", dtime=3, data_name="ECMWF")
    grd_fo_27 = meb.read_griddata_from_nc(path_fo_27, grid=grid1, time="2020072508", dtime=27, data_name="ECMWF")

    obs_array = grd_ob.values.squeeze()
    fst_array = grd_fo_03.values.squeeze()

    hold = make_spatialVx(obs_array, fst_array, loc=None,
                          fieldtype="Geometric Objects Pretending to be Precipitation",
                          units="mm/h", thresholds=[0.01, 50.01],
                          dataname="ICP Geometric Cases", obsname="geom000", modelname="geom001")
    look = locmeasures2d_spatial_vx(hold, k=[4, 0.975], alpha=0.1)
    print("hj")



import math

import numpy as np
import collections


def updateMatrix(mat):
    mat = np.array(mat)
    (m, n) = mat.shape
    dir = [(-1, 0), (0, 1), (1, 0), (0, -1)]
    queue = collections.deque()
    res = np.full((m, n), 0 + 0j)
    for r in range(m):
        for c in range(n):
            if mat[r][c] == 0:
                res[r][c] = 0 + 0j
                queue.append((r, c))
            else:
                res[r][c] = -1 - 1j

    while queue:
        r, c = queue.popleft()
        for dr, dc in dir:
            nr, nc = r + dr, c + dc
            if nr < 0 or nr >= m or nc < 0 or nc >= n or res[nr][nc] != -1 - 1j:
                continue
            res[nr][nc] = res[r][c] + dr + dc * 1j
            queue.append((nr, nc))
    new_res = np.zeros((m, n), dtype=float)
    for r in range(m):
        for c in range(n):
            new_res[r][c] = math.sqrt(np.real(res[r][c]) ** 2 + np.imag(res[r][c]) ** 2)
    return new_res



import numbers


def datagrabber_spatialVx(x, time_point=1, obs=1, model=1):
    if not isinstance(time_point, numbers.Number):
        raise Exception("datagrabber: invalid time.point argument.")
    if not isinstance(obs, numbers.Number):
        raise Exception("datagrabber: invalid obs argument.  Must be numeric.")
    if not isinstance(obs, numbers.Number):
        raise Exception("datagrabber: invalid model argument.  Must be numeric.")
    xdim = x["xdim"]
    nobs = x["nobs"]
    nf = x["nforecast"]
    vx = x['data'][0]
    if nobs > 1:
        vx = vx[obs]
    fcst = x['data'][1]
    if nf > 1:
        fcst = fcst[model]
    if len(xdim) == 3:
        vx = vx[:, :, time_point]
        fcst = fcst[:, :, time_point]
    out = {"X": vx, "Xhat": fcst}
    return out



def locmeasures2d_prep(input_object, k=None, alpha=0.1, bdconst=None, p=2):
    out = input_object
    out["alpha"] = alpha
    if bdconst is None:
        out["bdconst"] = float("inf")
    else:
        out["bdconst"] = bdconst
    out["p"] = p
    out["k"] = k
    return out



from . import yes_or_no
from . import multi_category
from . import continuous
from . import ensemble
from . import probability
from . import space
from . import vector
from . import weather_system
from .yes_or_no import *
from .multi_category import *
from .continuous import *
from .ensemble import *
from .probability import *
from .space import *
from .vector import *



import datetime
import copy
import numpy as np
import xarray as xr
import pandas as pd
import matplotlib.pyplot as plt
import os
import meteva
import urllib.request
import traceback
import math

def acc(grd_ob,grd_fo):
    grid0 = meteva.base.get_grid_of_data(grd_ob)
    z500_cli_path = meteva.base.z500_cli
    if not os.path.exists(z500_cli_path):
        url = "https://github.com/nmcdev/meteva/tree/master/meteva/resources/z500_cli.nc"
        try:
            print("开始从github下载500hPa高度场气候平均数据，请稍等")
            urllib.request.urlretrieve(url, filename=z500_cli_path)
            z500_cli = meteva.base.read_griddata_from_nc(z500_cli_path, grid=grid0)
            if z500_cli is None:
                raise ValueError("z500_cli 不能是None")
            print("下载500hPa高度场气候平均数据下载成功")
        except Exception as e:
            print("从github下载地形高度数据失败，请重试，或者手动从\n" + url + "\n下载文件并保存至\n" + z500_cli_path)
            print(e)
            if os.path.exists(z500_cli_path):
                os.remove(z500_cli_path)
            return
    else:
        z500_cli =  meteva.base.read_griddata_from_nc(z500_cli_path, grid=grid0)

    time_ob = meteva.base.all_type_time_to_datetime(grid0.gtime[0])
    if time_ob.hour ==0 or time_ob.hour == 12:
        time_ob_ut = time_ob
    else:
        time_ob_ut = time_ob - datetime.timedelta(hours=8)

    time_cli = datetime.datetime(2020, time_ob_ut.month, time_ob_ut.day, time_ob_ut.hour)
    grd_cli =  meteva.base.in_time_list(z500_cli, [time_cli])
    grd_obs = copy.deepcopy(grd_ob)
    grd_obs.values -= grd_cli.values
    grd_fst = copy.deepcopy(grd_fo)
    grd_fst.values -= grd_cli.values

    lats = grd_ob["lat"].values
    lons = grd_ob["lon"].values
    xx,yy = np.meshgrid(lons,lats)
    weight = np.cos(yy*math.pi / 180)
    grid0 = meteva.base.get_grid_of_data(grd_ob)
    grd_weight = meteva.base.grid_data(grid0,weight)


    acc_value =  meteva.method.corr(grd_obs.values,grd_fst.values,grd_weight.values)
    return acc_value



def acc_middle_z500(para_acc):
    z500_cli_path = meteva.base.z500_cli
    if not os.path.exists(z500_cli_path):
        url = "https://github.com/nmcdev/meteva/raw/master/meteva/resources/z500_cli.nc"
        try:
            print("开始从github下载地形高度数据，请稍等")
            urllib.request.urlretrieve(url, filename=z500_cli_path)
        except Exception as e:
            print("从github下载地形高度数据失败，请重试，或者手动从\n" + url + "\n下载文件并保存至\n" + z500_cli_path)
            print(e)
            return

    middle_result_path = para_acc["middle_result_path"]
    model_name_list =list(para_acc["fo_data"].keys()) # ["ECMWF","PANGU","FENGWU"]  #数据的名称
    if para_acc["step"] is not None:
        marker = meteva.perspact.get_grid_marker(para_acc["grid"],step=para_acc["step"])
    else:
        marker = None
    z500_cli =  meteva.base.read_griddata_from_nc(z500_cli_path,grid = para_acc["grid"])

    tmmsss_list = []
    #创建一个空的DataFrame，用来记录哪些中间量已经收集
    df1 = pd.DataFrame(data=None,columns=['level', 'time', 'dtime', 'member'])
    if os.path.exists(middle_result_path):
        tmmsss0 = pd.read_hdf(middle_result_path)
        tmmsss_list.append(tmmsss0)
        df1 = tmmsss0[["level","time","dtime","member"]]   #如果有文件就替换掉df1
        df1.drop_duplicates(keep="first", inplace=True)   # 用来记录哪些时次已经收集了中间量


    for model in model_name_list:
        df2 = df1.loc[df1['member'].isin([model])]
        time1 = para_acc["begin_time"]
        save_k = 0 # 用来在收集的过程间隔一段时间报错一下结果，避免程序错误时没有任何输出导致要重头再来
        para_fo1 = para_acc["fo_data"][model]
        dtime_list = np.arange(para_fo1["dtime"][0], para_fo1["dtime"][1] + 1, para_fo1["dtime"][2]).tolist()
        while time1 <=  para_acc["end_time"]:
            df3 =  meteva.base.in_time_list(df2,time1)
            if len(df3.index) ==len(dtime_list):
                print(meteva.base.get_path(model + "YYYY年MM月DD日HH时起报的检验中间量已存在",time1))
            else:
                time_ut = time1 - datetime.timedelta(hours=8)
                for dh in dtime_list:
                    df4 =  meteva.base.in_dtime_list(df3,dtime_list=dh)
                    if len(df4.index) ==0:
                        save_k += 1  #
                        time_ob = time1 + datetime.timedelta(hours=dh)
                        time_ob_ut = time_ob - datetime.timedelta(hours=8)
                        if para_acc["ob_data"]["time_type"] =="BT":
                            time_path = time_ob
                        else:
                            time_path = time_ob_ut
                        time_cli = datetime.datetime(2020,time_ob_ut.month,time_ob_ut.day,time_ob_ut.hour)
                        grd_cli =  meteva.base.in_time_list(z500_cli,[time_cli])

                        path0 =  meteva.base.get_path(para_acc["ob_data"]["dir_ob"], time_path)
                        grd_obs = para_acc["ob_data"]["read_method"](path0, grid=para_acc["grid"], time=time_ob, dtime=0, data_name="OBS",show = True,**para_acc["ob_data"]["read_para"])
                        if grd_obs is None: continue

                        grd_obs.values *= para_acc["ob_data"]["multiple"]
                        grd_obs.values -= grd_cli.values

                        if para_fo1["time_type"]=="BT":
                            path1 =  meteva.base.get_path(para_fo1["dir_fo"],time1,dh)
                        else:
                            path1 =  meteva.base.get_path(para_fo1["dir_fo"], time_ut, dh)

                        grd_fo = para_fo1["read_method"](path1, grid=para_acc["grid"], time=time1, dtime=dh, data_name=model,show = True,**para_fo1["read_para"])
                        if grd_fo is not None:
                            grd_fo.values *=para_fo1["multiple"]
                            grd_fo.values -= grd_cli.values
                            df_mid =  meteva.perspact.middle_df_grd(grd_obs, grd_fo, meteva.method.tmmsss, marker=marker)
                            tmmsss_list.append(df_mid)

            time1 = time1 + datetime.timedelta(hours=para_acc["time_step"])

            #当收集了超过200个时效的数据时就输出一次
            if save_k % 500==0:
                tmmss_all =  meteva.base.concat(tmmsss_list)
                #先删除文件在重新输出文件，避免文件大小膨胀
                if os.path.exists(middle_result_path):
                    os.remove(middle_result_path)
                tmmss_all.to_hdf(middle_result_path,"df")  #将结果输出到文件

    tmmss_all =  meteva.base.concat(tmmsss_list)
    # 先删除文件在重新输出文件，避免文件大小膨胀
    if os.path.exists(middle_result_path):
        os.remove(middle_result_path)
    tmmss_all.to_hdf(middle_result_path, "df")  # 将结果输出到文件
    print("中间量统计程序运行完毕")
    return


if __name__=="__main__":


    para_acc = {
        "begin_time": datetime.datetime(2022,5,1,8),  #起始时间（预报）
        "end_time":  datetime.datetime(2022,5,5,8), #结束时间（预报）
        "time_step":12,     #起报时间间隔
        "middle_result_path":  r"H:\test_data\output\method\space\acc\tmmsss_z500.h5",
        "grid" :  meteva.base.grid([70,140,0.25],[0,60,0.25]),  # 检验区域
        "step":2,  #masker间隔（单位：°）
        "ob_data": {
            "dir_ob":r"H:\test_data\input\mem\acc\ECMWF\z500\YYMMDDHH.TTT.nc",  # 实况场数据路径
            "read_method": meteva.base.io.read_griddata_from_nc,   #实况数据读取函数
            "read_para": {},  #实况数据读取参数
            "time_type": "BT",  #实况数据时间类型
            "multiple": 9.8,    #当文件数据的单位是位势米，通过乘以9.8转换成位势
        },
        "fo_data": {
            "ECMWF": {
                "dir_fo":r"H:\test_data\input\mem\acc\ECMWF\z500\YYMMDDHH.TTT.nc", #预报数据路径
                "dtime": [12, 72, 12],    #检验时效
                "read_method": meteva.base.io.read_griddata_from_nc,  #预报数据读取函数
                "read_para": {},    #预报数据读取参数
                "time_type": "BT",#预报数据时间类型是北京时，即08时起报
                "multiple":9.8, #当文件数据的单位是位势米，通过乘以9.8转换成位势
            },
            "NCEP": {
                "dir_fo": r"H:\test_data\input\mem\acc\NCEP\z500\YYMMDDHH.TTT.nc",  #预报数据路径
                "dtime": [12, 72, 12],  #检验时效
                "read_method": meteva.base.io.read_griddata_from_nc, #预报数据读取函数
                "read_para": {},#预报数据读取参数
                "time_type": "UT", #预报数据时间类型是世界时，即00时起报
                "multiple": 1, #当文件数据的单位是位势，则×1表示，不变
            },
        },
    }


    #acc_middle_z500(para_acc)
    df_mid = pd.read_hdf(para_acc["middle_result_path"])
    print(df_mid["member"])
    score,gdict = meteva.perspact.score_df(df_mid,meteva.method.corr,g = ["member","dtime"],plot = "line",save_path = r"H:\test_data\output\method\space\acc\acc_dtime.png",
                 xlabel = "时效（单位：小时）",ylabel = "ACC",title = "acc随时效的变化",grid = True,
                               height = 4,width = 7,sup_fontsize = 20)


    meteva.perspact.score_xy_df(df_mid,meteva.method.corr,g = ["member"],save_path = r"H:\test_data\output\method\space\acc\acc_xy.png",ncol = 2)



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 23:34:45 2022

@author: 1
"""
import numpy as npy
import pandas as pd
import copy
#import pyreadr
import sys
from scipy import optimize as optimize
from meteva.method.space.pphindcast.lib import hoods2dPrep, im, datagrabber, vxstats, thresholder
from meteva.method.space.pphindcast.lib import kernel2dsmooth, make_spatialVx
#sys.path.append(r'F:\Work\MODE')
#from tra_PracticallyPerfectHindcast.lib import hoods2dPrep
#from Submit2.baddeley_binary_image_metric.lib.im import im
#from Submit2.fuzzy_logic.lib import datagrabber, vxstats, thresholder, kernel2dsmooth
#from Submit2.baddeley_binary_image_metric.lib.make_spatialVx import make_spatialVx




def findthresh (p, Ix, sPx, binmat, which_score, subset = None):
    sIx = npy.zeros(binmat.shape)
    sIx[sPx >= p] = 1
    res = -vxstats.vxstats(sIx, Ix, which_stats=[which_score], subset=subset)[which_score]
    #print(res)
    return res

def Practically_Perfect_Hindcast(grd_ob,grd_fo,thresholds,compare = ">=",which_score = "ets",level = [1,3,5],
                                 smooth_fun ="hoods2dsmooth",show = False ):
    obs_array = grd_ob.values.squeeze()
    fst_array = grd_fo.values.squeeze()

    hold = make_spatialVx.make_spatialVx(obs_array, fst_array, loc=None,
                          fieldtype="Geometric Objects Pretending to be Precipitation",
                          units="mm/h", thresholds=thresholds,
                          dataname="ICP Geometric Cases", obsname="geom000", modelname="geom001")
    look_pphindcast2d = pphindcast2d(Object = hold,which_score=which_score,levels = level,verbose = show,rule= compare,
                                     smooth_fun = smooth_fun)
    return look_pphindcast2d


def pphindcast2d (Object, which_score = 'ets', time_point = 1, obs = 1, model = 1,
                  levels = None, max_n = None, smooth_fun = "hoods2dsmooth", 
                  smooth_params = None, rule = ">=", verbose = False):
    out = {}
    Object = hoods2dPrep.hoods2dPrep(input_object = Object, pe=None, levels=levels, max_n=max_n, 
                        smooth_fun="hoods2dsmooth", smooth_params=None)
    dat = datagrabber.datagrabber(Object)
    X = npy.array(dat['X'])
    Xhat = npy.array(dat['Xhat'])
    xdim = Object['xdim']
    Nxy = npy.prod(xdim)
    subset = Object['subset']
    thresholds = Object['thresholds'][1]
    levels = Object['levels']
    #q = npy.shape(thresholds[1])[0]
    q = npy.shape(thresholds)[0]
    l = len(levels)
    outmat = npy.zeros((l, q))
    Pthresh = npy.zeros((l, q))
    #out = {'attribute': Object, 'which_score': which_score}
    out = { 'which_score': which_score}
    binmat = npy.zeros((xdim[0], xdim[1]))
    for threshold in range(q):
        #print(threshold)
        dat2_X = thresholder.thresholder(x=X, th=thresholds[threshold], func_type='binary', rule=">=", replace_with=0)
        dat2_Xhat = thresholder.thresholder(x=Xhat, th=thresholds[threshold], func_type='binary', rule=">", replace_with=0)
        Ix = dat2_X
        Iy = dat2_Xhat
        for level in range(l):
            #print(level)
            Wlvl = kernel2dsmooth.kernel2dsmooth(Ix, kernel_type='boxcar', n = levels[level], setup=True)
            sPy = kernel2dsmooth.kernel2dsmooth(Iy, kernel_type='boxcar', n = levels[level],
                                 W = Wlvl, xdim = xdim, Nxy = Nxy)
            sPx = kernel2dsmooth.kernel2dsmooth(Ix, kernel_type='boxcar', n = levels[level],
                                 W = Wlvl, xdim = xdim, Nxy = Nxy)
            Pthresh[level, threshold] = optimize.minimize_scalar(findthresh, 
                                                                 args=(Ix, sPx, binmat, which_score),
                                                                 bounds = (0,1), 
                                                                 method = "bounded")['x']

            #Pthresh[level, threshold] = optimize.minimize(findthresh,0,(Ix, sPx, binmat, which_score),"L-BFGS-B")
           
            
            
            if verbose:
                print("Pthresh = ", Pthresh[level, threshold],
                      "for obs threshold no. = ", threshold , "and level = ",
                      levels[level], "\n")
            sIy = thresholder.thresholder(x = sPy, th = Pthresh[level, threshold], 
                                          func_type = "binary", rule = rule)
            outmat[level, threshold] = vxstats.vxstats(x = sIy, xhat = Ix, 
                                                       which_stats = [which_score])[which_score]
    out['value'] = outmat
    out['Pthresh'] = Pthresh
    out['time_point'] = time_point
    out['model_num'] = model
    out['class'] = "pphindcast"
    
    
    return out


if __name__ == '__main__':
    # geom000 = pyreadr.read_r(r'F:\Work\MODE\Submit2\baddeley_binary_image_metric\data\geom000.Rdata')['geom000']
    # geom001 = pyreadr.read_r(r'F:\Work\MODE\Submit2\baddeley_binary_image_metric\data\geom001.Rdata')['geom001']
    # ICPg240Locs = pyreadr.read_r(r'F:\Work\MODE\Submit2\baddeley_binary_image_metric\data\ICPg240Locs.Rdata')['ICPg240Locs']

    import meteva.base as meb
    grid1 = meb.grid([100, 120, 0.05], [24, 40, 0.05])
    path_ob = r'H:\test_data\input\mem\mode\ob\rain03\20072611.000.nc'
    path_fo = r'H:\test_data\input\mem\mode\ec\rain03\20072608.003.nc'
    grd_ob = meb.read_griddata_from_nc(path_ob, grid=grid1, time="2020072611", dtime=0, data_name="OBS")
    grd_fo = meb.read_griddata_from_nc(path_fo, grid=grid1, time="2020072608", dtime=3, data_name="ECMWF")

    res = Practically_Perfect_Hindcast(grd_ob,grd_fo,thresholds=[0.01, 50.01])


    '''
    Object = hold
    which_score = 'ets'
    time_point = 1
    obs = 1
    model = 1
    levels = npy.array([1,3,65])
    max_n = None
    smooth_fun = "hoods2dsmooth"
    smooth_params = None
    rule = ">="
    verbose = False
    '''
    
    

    
    
    


# -*-coding:utf-8-*-

import math
import sys
from scipy.spatial import cKDTree
from meteva.method.space.baddeley_binary_image_metric.lib.distmap import *


def cv2_distanceTransform(array_01):
    #计算每个格点到目标的最近距离
    nx = array_01.shape[1]
    ny = array_01.shape[0]
    index = np.where(array_01 ==0)
    xy = np.array(index).T
    tree = cKDTree(xy)
    xy_g = np.array(np.meshgrid(np.arange(ny),np.arange(nx))).T
    d,_ = tree.query(xy_g, k=1)
    d_g = d.reshape(ny,nx)
    return d_g


def deltametric(a, b, p=2, c=float('inf')):
    import cv2
    if p == float('inf') or (type(p).__name__ != "str" and str(p).isnumeric() and p > 0):
        window = boundingbox(a, b)
        a = rebound(a, window)
        b = rebound(b, window)
        # dA = ndimage.morphology.distance_transform_edt(~(a['m'] == 1) + 0)
        # dB = ndimage.morphology.distance_transform_edt(~(b['m'] == 1) + 0)
        d_a = cv2.distanceTransform(np.array(~(a['m'] == 1) + 0, np.uint8), cv2.DIST_L2, 3, dstType=cv2.CV_32F)
        d_b = cv2.distanceTransform(np.array(~(b['m'] == 1) + 0, np.uint8), cv2.DIST_L2, 3, dstType=cv2.CV_32F)
        if not math.isinf(c):
            d_a = np.minimum(d_a, c)
            d_b = np.minimum(d_b, c)
        if math.isinf(p):
            z = np.abs(d_a - d_b)
            delta = np.max(z)
        else:
            z = np.abs(d_a - d_b) ** p
            i_z = np.mean(z)
            delta = i_z ** (1.0 / p)
        return delta
    else:
        print("p类型错误")
        raise Exception("p类型错误")



def as_unitname(s):
    n = 0
    if s is not None:
        n = len(s)
    if n > 3:
        raise Exception("Unit name should be a character string,", "or a vector/list of 2 character strings,",
                        "or a list(character, character, numeric)")
    # out = switch(n + 1, makeunitname(), makeunitname(s[[1]], s[[1]]), makeunitname(s[[1]], s[[2]]), makeunitname(s[[1]], s[[2]], s[[3]]))
    out = None
    if n == 0:
        out = makeunitname()
    elif n == 1:
        out = makeunitname(s[[0]], s[[0]])
    elif n == 2:
        out = makeunitname(s[[0]], s[[1]])
    elif n == 3:
        out = makeunitname(s[[0]], s[[1]], s[[2]])
    return out


def makeunitname(sing="unit", plur="units", mul=1):
    out = {
        'singular': sing,
        'plural': plur,
        'multiplier': mul
    }
    return out



import math
import numpy as np


def kernel2dmeitsjer(func_type="gauss", **args):
    theta = args
    if 'h' in theta and theta['h'] is not None:
        h = theta['h']
    elif 'nx' in theta and 'ny' in theta and theta['nx'] is not None and theta['ny'] is not None:
        nx = theta['nx']
        ny = theta['ny']
        if 'a' in theta and theta['a'] is not None:
            a = theta['a']
        else:
            a = 1
        xgrid = np.array([val for val in range(1, nx + 1) for i in range(ny)]).reshape(nx, ny)
        ygrid = np.array([val for val in range(1, ny + 1) for i in range(nx)]).reshape(nx, ny)
        xcen = nx / 2
        ycen = ny / 2
        h = (xgrid - xcen) ** 2 + (ygrid - ycen) ** 2

    if 'sigma' in theta and theta['sigma'] is not None:
        sigma = theta['sigma']
        sigma2 = sigma ** 2
    if func_type == "average":
        out = np.ones([theta['n'], theta['n']]) * (1 / (theta['nx'] * theta['ny']))
    elif func_type == "boxcar":
        out = np.ones([theta['n'], theta['n']]) * (1 / (theta['n'] ** 2))
    elif func_type == "cauchy":
        out = 1 / (1 + h / sigma)
    elif func_type == "disk":
        r = theta['r']
        r2 = r ** 2
        rint = np.ceil(r - 0.5)
        diskg = [n for n in range(-rint, rint + 1)]
        n = len(diskg)
        x = np.array([val for val in diskg for i in range(n)]).reshape(n, n)
        y = np.transpose(x)
        xy_absmax = np.maximum(np.abs(x))
        xy_absmin = np.minimum(np.abs(x))
        tmp1 = tmp2 = hold = np.zeros(n, n)
        val = (xy_absmax + 0.5) ** 2 + (xy_absmin - 0.5) ** 2
        id1 = r2 < val
        id2 = r2 >= val
        tmp1[id1] = (xy_absmin[id1] - 0.5)
        tmp1[id2] = np.sqrt(r ** 2 - (xy_absmax[id2] + 0.5) ** 2)
        val = (xy_absmax - 0.5) ** 2 + (xy_absmin + 0.5) ** 2
        id1 = r2 > val
        id2 = r2 <= val
        tmp2[id1] = xy_absmin[id1] + 0.5
        tmp2[id2] = np.sqrt(r2 - (xy_absmax[id2] - 0.5) ** 2)
        val1 = (xy_absmax + 0.5) ** 2 + (xy_absmin + 0.5) ** 2
        hold_id1 = ((r2 < val1) & (r2 > (xy_absmax - 0.5) ** 2 + (xy_absmin - 0.5) ** 2))
        hold_id2 = ((xy_absmin == 0) & (xy_absmax - 0.5 < r) & (xy_absmax + 0.5 >= r))
        hold_id = hold_id1 | hold_id2
        hold[hold_id] = r2 * (0.5 * (math.asin(tmp2[hold_id] / r) - math.asin(tmp1[hold_id] / r)) + 0.25 * (
                    np.sin(2 * math.asin(tmp2[hold_id] / r)) - np.sin(2 * math.asin(tmp1[hold_id] / r)))) - (
                                    xy_absmax[hold_id] - 0.5) * (tmp2[hold_id] - tmp1[hold_id]) + (
                                    tmp1[hold_id] - xy_absmin[hold_id] + 0.5)
        hold[np.isnan(hold)] = 0
        hold = hold + (val1 < r2)
        hold[rint + 1, rint + 1] = min(np.pi * r2, np.pi / 2)
        rc2 = rint - 0.5
        if rint > 0 and r > rc2 and r2 < rc2 ** 2 + 0.25:
            tmp1 = np.sqrt(r2 - rc2 ** 2)
            tmp1n = tmp1 / r
            hold0 = 2 * (r ** 2 * (0.5 * math.asin(tmp1n) + 0.25 * np.sin(2 * math.asin(tmp1n))) - tmp1 * (rint - 0.5))
            hold[2 * rint + 1, rint + 1] = hold0
            hold[rint + 1, 2 * rint + 1] = hold0
            hold[rint + 1, 1] = hold0
            hold[1, rint + 1] = hold0
            hold[2 * rint, rint + 1] = hold[2 * rint, rint + 1] - hold0
            hold[rint + 1, 2 * rint] = hold[rint + 1, 2 * rint] - hold0
            hold[rint + 1, 2] = hold[rint + 1, 2] - hold0
            hold[2, rint + 1] = hold[2, rint + 1] - hold0

        hold[rint + 1, rint + 1] = min(hold[rint + 1, rint + 1], 1)
        Sn = np.sum(hold)
        out = hold / Sn
    elif func_type == "epanechnikov":
        out = np.zeros(nx, ny)
        out[h <= 1] = 3 / 4 * (1 - h[h <= 1] / sigma2)
    elif func_type == "exponential":
        if theta['a'] is not None:
            a = theta['a']
        else:
            a = 1
        h = np.sqrt(h)
        out = a * math.exp(-h / (2 * sigma2))
    elif func_type == "gauss":
        out = (1 / (2 * np.pi * sigma2)) * math.exp(-h / (2 * sigma2))
    elif func_type == "laplacian" or func_type == "unsharp":
        if theta['a'] is not None:
            a = theta['alpha']
        else:
            a = 0
        a = max(0, min(a, 1))
        o1 = a / (a + 1)
        o2 = (1 - a) / (1 + a)
        out = np.vstack([o1, o2, o1], [o2, -4 / (a + 1), o2], [o1, o2, o1])
        if func_type == "unsharp":
            out = np.vstack([0] * 3, [0, 1, 0], [0] * 3) - out
    elif func_type == "LoG":
        out = (h - 2 * sigma2) * np.exp(-h / (2 * sigma2)) / (sigma2 ** 2)
        out = out - np.mean(out)
    elif func_type == "minvar":
        out = np.zeros(nx, ny)
        out[h <= 1] = 3 / 8 * (3 - 5 * h[h <= 1] / sigma2)
    elif func_type == "multiquad":
        a2 = (theta['a']) ** 2
        if theta['inverse'] is None:
            inverse = False
        else:
            inverse = theta['inverse']
        out = np.sqrt((h + a2))
        if inverse:
            out = 1 / out
    elif func_type == "power":
        p = theta['p']
        h = np.sqrt(h)
        if theta['do.log'] is None:
            do_log = False
        else:
            do_log = theta['do.log']
        if ~do_log:
            out = -h ** p
        else:
            out = -np.log(h ** p + 1)
    elif func_type == "prewitt":
        out = np.vstack([1] * 3, [0] * 3, [-1] * 3)
        if theta['transpose'] is not None:
            if theta['transpose']:
                out = np.transpose(out)
    elif func_type == "radial":
        a = theta['a']
        m = theta['m']
        d = theta['d']
        if d % 2 == 0:
            out = a * h ** (2 * m - d) * np.log(h)
        else:
            out = a * h ** (2 * m - d)
        out[np.isnan(out)] = 0
    elif func_type == "ratquad":
        a = theta['a']
        out = 1 - h / (h + a)
    elif func_type == "sobel":
        out = np.vstack([1, 2, 1], [0] * 3, [-1, -2, -1])
        if theta['transpose'] is not None:
            if theta['transpose']:
                out = np.transpose(out)
    elif func_type == "student":
        p = theta['p']
        h = np.sqrt(h)
        out = 1 / (1 + h ** p)
    elif func_type == "wave":
        phi = theta['phi']
        h = np.sqrt(h)
        out = (phi / h) * np.sin(h / phi)
        out[np.isnan(out)] = 0

    return out



from meteva.base import *
from meteva.method import *
from meteva.product.program.fun import *
import pandas as pd
import numpy as np


def score(sta_ob_and_fos0,method,s = None,g = None,gll = None,group_name_list = None,plot = None,
          vmax = None,vmin = None,bar_width = None,save_path = None,show = False,dpi = 300,title = "",
          excel_path = None,drop_g_column = False,spasify_xticks = None,keep_ob_IV = False,**kwargs):


    if s is not None:
        if g is not None:
            if g == "last_range" or g == "last_step":
                s["drop_last"] = False
            else:
                s["drop_last"] = True
    sta_ob_and_fos = sele_by_dict(sta_ob_and_fos0, s)
    if not keep_ob_IV:
        index = np.where(sta_ob_and_fos.iloc[:,6].values != meteva.base.IV)
        sta_ob_and_fos = sta_ob_and_fos.iloc[index[0],:]
    if(len(sta_ob_and_fos.index) == 0):
        msg = "there is no data to verify"
        print(msg)
        return None,msg

    if type(method) == str:
        method =  globals().get(method)


    if method == meteva.method.fss_time:
        if g == "dtime":
            msg = "fss_time 检验时，参数group_by不能选择dtime"
            print(msg)
            return

    # get method_args and plot_args


    iv_in_fo = meteva.base.IV in sta_ob_and_fos.iloc[:,7:].values

    method_args = {}
    plot_args = {}
    method_para_list = method.__code__.co_varnames
    plot_mehod = None
    if plot == "bar":
        plot_mehod = meteva.base.plot_tools.bar
    elif plot == "line" or plot == "plot":
        plot_mehod = meteva.base.plot_tools.plot

    plot_para_list = []
    if plot_mehod is not None:
        plot_para_list = plot_mehod.__code__.co_varnames

    for key in kwargs.keys():
        if key in method_para_list and key not in plot_para_list:
            method_args[key] = kwargs[key]
        elif key not in method_para_list and key  in plot_para_list:
            plot_args[key] = kwargs[key]
        elif key  in method_para_list and key  in plot_para_list:
            msg = method.__name__ + " and " + plot_mehod.__name__ + " have same args:" + key
            print(msg)
            return None,msg
        else:
            if plot_mehod is None:
                msg = key + " is not args of " + method.__name__
            else:
                msg = key + " is not args of " + method.__name__ + " or " + plot_mehod.__name__
            print(msg)
            return None,msg

    sta_ob_and_fos_list,group_list_list1 = group(sta_ob_and_fos,g,gll,drop_g_column)
    group_num = len(sta_ob_and_fos_list)
    if group_num ==0:
        msg = "数据中不包含指定方式的分类结果"
        print(msg)
        return  None,msg

    data_name = meteva.base.get_stadata_names(sta_ob_and_fos_list[0])
    if method.__name__.find("ob_fo")>=0:
        fo_name = data_name
    elif method.__name__.find("_uv")>=0:
        fo_name = []
        for i in range(2,len(data_name),2):
            strs = data_name[i]
            strs = strs.replace("u_","")
            fo_name.append(strs)
    elif method.__name__.find("distance")>=0:
        fo_name = []
        for i in range(2,len(data_name),2):
            strs = data_name[i]
            strs = strs.replace("lon_","")
            fo_name.append(strs)
    else:
        ensemble_score_method = [meteva.method.cr,meteva.method.crps,variance_divide_by_mse]
        if method in ensemble_score_method:
            fo_name = [""]
        elif method ==meteva.method.variance_mse:
            fo_name = ["variance","MSE"]
        elif method ==meteva.method.std_rmse:
            fo_name = ["STD","RMSE"]
        else:
            fo_name = data_name[1:]

    fo_num = len(fo_name)


    #等级参数的确定
    if "grade_list" not in kwargs.keys():
        mutil_list = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                      meteva.method.mr_multi, meteva.method.far_multi]
        if method in mutil_list:
        # 如果是多分类检验，但又没有设置分级方法，就需要从数据中获得全局的种类
            values = sta_ob_and_fos.iloc[:,6:].values.flatten()
            index_list = list(set(values))
            if len(index_list) > 30:
                msg = "自动识别的样本类别超过30种，判断样本为连续型变量，grade_list不能缺省"
                print(msg)
                return None,msg
            index_list.sort()
            grade_list = []
            for i in range(len(index_list)-1):
                grade_list.append((index_list[i]+index_list[i+1])/2)
            kwargs["grade_list"] = grade_list

    if "grade_list" in kwargs.keys():
        grades = kwargs["grade_list"]
        grade_names = []
        mutil_list1 = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                       meteva.method.mr_multi, meteva.method.far_multi,
                       meteva.method.ts_grade, meteva.method.bias_grade, meteva.method.ets_grade,
                       meteva.method.mr_grade, meteva.method.far_grade]
        mutil_list2 = [meteva.method.accuracy,meteva.method.hk,meteva.method.hss]
        if method in mutil_list1:
            grade_names = ["<" +  str(grades[0])]
            for i in range(len(grades)-1):
                grade_names.append("["+str(grades[i]) + ","+str(grades[i+1])+")")
            grade_names.append(">=" + str(grades[-1]))
        elif method in mutil_list2:
            grade_names = ["0"]
        else:
            for i in range(len(grades)):
                grade_names.append(grades[i])
    else:
        grade_names = ["0"]
    grade_num = len(grade_names)

    if method == meteva.method.fss_time:
        #统计dtime的集合
        dtime_list = list(set(sta_ob_and_fos["dtime"].values.tolist()))
        dtime_list.sort()
        ndtime = len(dtime_list)
        result= []
        for sta_ob_and_fo in sta_ob_and_fos_list:
            # 将观测和预报数据重新整理成fss_time所需格式
            ob = in_member_list(sta_ob_and_fo,[data_name[0]])
            ob_dtimes = None
            for k in range(ndtime):
                dtimek = dtime_list[k]
                sta_obk = in_dtime_list(ob,[dtimek])
                set_stadata_names(sta_obk,[data_name[0]+ str(dtimek)])
                ob_dtimes = combine_on_leve_time_id(ob_dtimes,sta_obk)
            result1 = []
            #print(ob_dtimes)
            ob_array = ob_dtimes.values[:,6:]
            for j in range(fo_num):
                fo = in_member_list(sta_ob_and_fo, [data_name[j+1]])
                fo_dtimes = None
                for k in range(ndtime):
                    dtimek = dtime_list[k]
                    sta_fok = in_dtime_list(fo, [dtimek])
                    set_stadata_names(sta_fok, [data_name[j+1] + str(dtimek)])
                    fo_dtimes = combine_on_leve_time_id(fo_dtimes, sta_fok)
                fo_array = fo_dtimes.values[:,6:]

                #调用检验程序
                result2 = fss_time(ob_array, fo_array, **kwargs)
                result1.append(result2)
            result.append(result1)
        result = np.array(result) #将数据转换成数组
        result = result.squeeze()

    else:
        result_list = []
        for i in range(group_num):
            sta = sta_ob_and_fos_list[i]

            not_all_iv = [True]
            if method.__name__.find("_uv") >= 0 or method.__name__.find("distance") >= 0:
                col_step = 2
                valid_index = [0,1]
            else:
                col_step = 1
                valid_index = [0]
            if iv_in_fo:
                len_ = len(sta.columns)
                for nv in range(6+col_step,len_,col_step):
                    not_all_iv1 = np.any(sta.iloc[:,nv].values != meteva.base.IV)
                    not_all_iv.append(not_all_iv1)
                    if not_all_iv1:
                        valid_index.append(nv-6)
                        if col_step == 2:valid_index.append(nv - 5)
                #print(valid_index)
                sta = meteva.base.in_member_list(sta,member_list=valid_index,name_or_index ="index")
                sta = meteva.base.not_IV(sta)

            data_name = meteva.base.get_stadata_names(sta)
            if(len(sta.index) == 0):
               result1 = np.zeros((fo_num,grade_num)).squeeze()
            else:
                if method.__name__.find("_uv")>=0:
                    u_ob = sta[data_name[0]].values
                    v_ob = sta[data_name[1]].values
                    u_fo = sta[data_name[2::2]].values.T
                    v_fo = sta[data_name[3::2]].values.T
                    result1 = method(u_ob,u_fo,v_ob,v_fo,**method_args)
                elif method.__name__.find("distance")>=0:
                    x_ob = sta[data_name[0]].values
                    y_ob = sta[data_name[1]].values
                    x_fo = sta[data_name[2::2]].values.T
                    y_fo = sta[data_name[3::2]].values.T
                    result1 = method(x_ob,y_ob,x_fo,y_fo,**method_args)
                else:
                    ob = sta[data_name[0]].values
                    fo = sta[data_name[1:]].values.T
                    #if ob.size>0:
                    result1 = method(ob,fo,**method_args)
                    # else:
                    #     result1= np.ones(fo_num) * meteva.base.IV

            #if len(result1.shape)==2: result1 = result1.squeeze()
            if iv_in_fo:
                di = len(not_all_iv) - fo_num
                shape_ = result1.shape
                if len(shape_)==2:
                    result_with_iv = np.ones((fo_num,shape_[1])) * meteva.base.IV
                    kiv = 0
                    for nv in range(fo_num):
                        if not_all_iv[nv+di]:
                            result_with_iv[nv,:] = result1[kiv,:]
                            kiv += 1
                else:
                    result_with_iv = np.ones(fo_num) * meteva.base.IV
                    kiv = 0
                    for nv in range(fo_num):
                        if not_all_iv[nv+di]:
                            result_with_iv[nv] = result1[kiv]
                            kiv += 1
                result1 = result_with_iv.squeeze()
            result_list.append(result1)
        result = np.array(result_list)

    if plot is not None or excel_path is not None:

        name_list_dict = {}
        if g is None:
            group_dict_name = "group_name"
        else:
            group_dict_name = g
        #设置分组名称
        if group_name_list is not None:
            if group_num == len(group_name_list):
                name_list_dict[group_dict_name] =group_name_list
            else:
                print("group_name_list参数中包含的分组名称个数和实际分组个数不匹配")
        else:
            if not isinstance(group_list_list1,list):
                group_list_list1 = [group_list_list1]

            if (group_dict_name == "time" or group_dict_name == "ob_time" or group_dict_name == "day" or group_dict_name == "ob_day")and gll is None:
                name_list_dict[group_dict_name] = group_list_list1
            else:
                name_list_dict[group_dict_name] = get_group_name(group_list_list1)

        #设置成员名称

        name_list_dict["member"] = fo_name
        if fo_num==0:
            fo_num = 1
            name_list_dict["member"] = ["OBS"]
        result_plot = result.reshape((group_num,fo_num,grade_num))
        #设置等级名称
        name_list_dict["grade"] = grade_names
        keys = list(name_list_dict.keys())
        if fo_num ==1:
            if grade_num > 1:
                if group_num>1:
                    legend = keys[2]
                    axis = keys[0]
                else:
                    axis = keys[2]
                    legend = keys[1]
            else:
                axis = keys[0]
                legend = keys[1]
        else:
            if group_num == 1 :
                if grade_num >1:
                    legend = keys[1]
                    axis = keys[2]
                else:
                    legend = keys[2]
                    axis = keys[1]
            else:
                legend = keys[1]
                axis = keys[0]
        if "ylabel" not in plot_args.keys():
            plot_args["ylabel"] = method.__name__



        bigthan0_method = [meteva.method.ts,meteva.method.ob_fo_hr,meteva.method.ob_fo_std,meteva.method.ts_multi,
                           meteva.method.s,meteva.method.pc_of_sun_rain,meteva.method.bias_multi,meteva.method.bias,
                           meteva.method.pc,meteva.method.mr,meteva.method.far,meteva.method.tc,
                           meteva.method.roc_auc,meteva.method.r,meteva.method.sr,meteva.method.cr,meteva.method.pod,
                           meteva.method.pofd,meteva.method.mse,meteva.method.rmse,meteva.method.mae]

        if vmin is None:
            if method in bigthan0_method:
                vmin = 0
        if "hline" not in plot_args.keys():
            if method in [meteva.method.me]:
                plot_args["hline"] = 0
            elif method in [meteva.method.bias]:
                plot_args["hline"] = 1

        if plot is not None:
            if plot =="bar":
                meteva.base.plot_tools.bar(result_plot,name_list_dict,legend=legend,axis = axis,vmin =vmin,vmax = vmax,bar_width=bar_width,save_path=save_path,show=show,dpi =dpi,title = title,spasify_xticks=spasify_xticks,**plot_args)
            else:
                meteva.base.plot_tools.plot(result_plot,name_list_dict,legend=legend,axis = axis,vmin =vmin,vmax = vmax,save_path=save_path,show=show,dpi = dpi,title= title,spasify_xticks=spasify_xticks,**plot_args)
        if excel_path is not None:
            meteva.base.write_array_to_excel(result_plot,excel_path,name_list_dict,index= axis,columns=legend)

    if result.size != 1:
        result = result.squeeze()
    return result,group_list_list1



def score_id(sta_ob_and_fos0,method,s = None,g = None,gll = None,group_name_list = None,plot = "scatter",save_dir = None,save_path = None,show = False,
             add_county_line = False,map_extend= None,print_max=0,print_min=0,dpi = 300,title = None,sort_by = None,subplot = "member",ncol = None,
             **kwargs):

    if s is not None:
        if g is not None:
            if g == "last_range" or g == "last_step":
                s["drop_last"] = False
            else:
                s["drop_last"] = True

    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0,s = s)
    if(len(sta_ob_and_fos1.index) == 0):
        print("there is no data to verify")
        return

    sta_ob_and_fos_list, gll1 = meteva.base.group(sta_ob_and_fos1, g = g, gll = gll)

    if "grade_list" in kwargs.keys():
        grades = kwargs["grade_list"]
        grade_names = []
        mutil_list1 = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                       meteva.method.mr_multi, meteva.method.far_multi,
                       meteva.method.ts_grade, meteva.method.bias_grade, meteva.method.ets_grade,
                       meteva.method.mr_grade, meteva.method.far_grade]
        mutil_list2 = [meteva.method.accuracy,meteva.method.hk,meteva.method.hss]
        if method in mutil_list1:
            grade_names = ["<" +  str(grades[0])]
            for i in range(len(grades)-1):
                grade_names.append("["+str(grades[i]) + ","+str(grades[i+1])+")")
            grade_names.append(">=" + str(grades[-1]))
        elif method in mutil_list2:
            grade_names = ["0"]
        else:
            for i in range(len(grades)):
                grade_names.append(grades[i])
    else:
        grade_names = ["0"]
    grade_num = len(grade_names)

    g_num = len(sta_ob_and_fos_list)
    if(g_num == 1):gll1 = [None]
    if group_name_list is None:
        group_name_list = meteva.product.program.get_group_name(gll1)


    data_names = meteva.base.get_stadata_names(sta_ob_and_fos_list[0])
    ensemble_score_method = [meteva.method.cr,meteva.method.crps, variance_divide_by_mse]
    if method.__name__.find("ob_fo")>=0:
        fo_name = data_names

    elif method.__name__.find("_uv")>=0:
        fo_name = []
        for i in range(2,len(data_names),2):
            strs = data_names[i]
            strs = strs.replace("u_","")
            fo_name.append(strs)

    elif method.__name__ == "sample_count":
        fo_name = [data_names[0]]
    elif method in ensemble_score_method:
        fo_name = [""]
    elif method == meteva.method.variance_mse:
        fo_name = ["variance","MSE"]
    elif method ==meteva.method.std_rmse:
        fo_name = ["STD","RMSE"]
    else:
        fo_name = data_names[1:]
    fo_num = len(fo_name)

    if title is not None:
        if isinstance(title, list):
            if fo_num * g_num * grade_num != len(title):
                print("手动设置的title数目和要绘制的图形数目不一致")
                return

    if save_path is not None:
        if isinstance(save_path, str):
            save_path = [save_path]
        if "subplot" is not None:
            if g_num * grade_num != len(save_path):
                print("手动设置的save_path数目和要绘制的图形数目不一致")
                return
        else:
            if fo_num * g_num * grade_num != len(save_path):
                print("手动设置的save_path数目和要绘制的图形数目不一致")
                return

    # get method_args and plot_args
    method_args = {}
    plot_args = {}
    method_para_list = method.__code__.co_varnames
    plot_mehod = None
    if plot == "scatter":
        plot_mehod = meteva.base.plot_tools.scatter_sta
    elif plot == "contourf":
        plot_mehod = meteva.base.plot_tools.contourf_2d_grid

    plot_para_list = []
    if plot_mehod is not None:
        plot_para_list = plot_mehod.__code__.co_varnames

    for key in kwargs.keys():
        if key in method_para_list and key not in plot_para_list:
            method_args[key] = kwargs[key]
        elif key not in method_para_list and key  in plot_para_list:
            plot_args[key] = kwargs[key]
        elif key  in method_para_list and key  in plot_para_list:
            print(method.__name__ + " and " + plot_mehod.__name__ + " have same args:" + key)
            return
        else:
            print(key + " is not args of " + method.__name__ + " or " + plot_mehod.__name__)
            return

    if "cmap" not in plot_args.keys() and "clevs" not in plot_args:
        if method == meteva.method.me:
            plot_args["cmap"] = meteva.base.cmaps.me


    result_all = []
    for k in range(g_num):
        g_id = "id"
        result,id_list= score(sta_ob_and_fos_list[k],method,g = g_id,**method_args)
        station = sta_ob_and_fos_list[k].drop_duplicates(['id'],inplace=False)
        station.iloc[:, 1] = station.iloc[0, 1]
        station.iloc[:, 2] = station.iloc[0, 2]
        station.iloc[:, 0] = station.iloc[0, 0]
        station1 = meteva.base.in_id_list(station,id_list)
        id_s = pd.Series(id_list)
        id_s.name = "id"
        sta_merge = pd.merge(id_s,station1, on='id', how='left')
        sta_merge = meteva.base.sta_data(sta_merge)
        coord_names = meteva.base.get_coord_names()
        if len(result.shape) == 1:
            #没有等级，只有一个预报成员
            coord_names.append(data_names[-1])
            sta_result = sta_merge.loc[:,coord_names]
            sta_result.iloc[:,-1] = result[:]
            sta_result = [sta_result]
        elif len(result.shape) == 2:
            if len(fo_name) >1:
                #没有等级，但有多个预报成员
                if method.__name__.find("_uv") >= 0:
                    #member_num = result.shape[1]
                    #coord_names.extend(fo_name)
                    sta_result = sta_merge.loc[:, coord_names]
                    for ff in range(len(fo_name)):
                        sta_result[fo_name[ff]] = result[:,ff]
                    sta_result = [sta_result]
                elif method == meteva.method.variance_mse:
                    member_num = result.shape[1]
                    #coord_names.extend(fo_name)
                    sta_result = sta_merge.loc[:, coord_names]
                    sta_result["variance"] = result[:, 0]
                    sta_result["MSE"] = result[:, 1]
                    sta_result = [sta_result]
                elif method == meteva.method.std_rmse:
                    member_num = result.shape[1]
                    #coord_names.extend(fo_name)
                    sta_result = sta_merge.loc[:, coord_names]
                    sta_result["STD"] = result[:, 0]
                    sta_result["RMSE"] = result[:, 1]
                    sta_result = [sta_result]
                else:
                    member_num = result.shape[1]
                    coord_names.extend(fo_name)
                    sta_result = sta_merge.loc[:, coord_names]
                    sta_result.iloc[:, -member_num:] = result[:,:]
                    sta_result = [sta_result]
            else:
                #有多个等级，但只有一个预报成员
                sta_result = []
                coord_names.append(data_names[-1])
                for i in range(result.shape[1]):
                    sta_result1 = sta_merge.loc[:, coord_names]
                    sta_result1.iloc[:, -1] = result[:,i]
                    sta_result.append(sta_result1)
        else:
            #有多个等级，同时有多个预报成员
            sta_result = []
            coord_names.extend(fo_name)
            member_num = result.shape[1]
            for i in range(result.shape[2]):
                sta_result1 = sta_merge.loc[:, coord_names]
                sta_result1.iloc[:, -member_num:] = result[:,:,i]
                sta_result.append(sta_result1)
        sup_title = None
        if "sup_title" in plot_args.keys():
            sup_title = copy.deepcopy(plot_args["sup_title"])

        for i in range(len(sta_result)):
            sta_result1 = sta_result[i]
            if "var_name" in sta_ob_and_fos0.attrs.keys():
                sta_result1.attrs["var_name"] =sta_ob_and_fos0.attrs["var_name"]
            else:
                sta_result1.attrs["var_name"] = ""
            sta_result1.attrs["data_source"] = "meteva." + method.__name__
            if plot is not None:

                if isinstance(title, list):
                    kk = k * grade_num + i
                    title1_list = title[kk * fo_num: (kk + 1) * fo_num]
                else:
                    title1_list = []
                    for ii in range(fo_num):
                        if g !="time":
                            if title is not None:
                                title1 =meteva.product.program.get_title_from_dict(title, s, g, group_name_list[k],
                                                                                     fo_name[ii])
                            else:
                                title1 = meteva.product.program.get_title_from_dict("", s, g,
                                                                                    group_name_list[k], fo_name[ii])
                        else:
                            title1 = ""
                        if grade_num>1:
                            title1 += "(grade_" + str(grade_names[i])+")"
                        title1_list.append(title1)


                if plot == "scatter" or plot =="contourf":
                    if "sup_title" not in plot_args.keys():
                        plot_args["sup_title"] = ""
                    elif isinstance(sup_title,list):
                        plot_args["sup_title"] = sup_title[i]
                    save_path1 = None
                    if save_path is None:
                        if save_dir is None:
                            show = True
                        else:
                            save_path1 = []
                            for i in range(len(title1_list)):
                                fileName = title1_list[i].replace("\n", "").replace(":", "")
                                save_path1.append(save_dir + "/" + fileName + ".png")
                    elif subplot=="member":
                        save_path1 = save_path[i]
                    else:
                        save_path1 = save_path[k * fo_num: (k + 1) * fo_num]

                    if plot=="scatter":
                        if "fix_size" not in plot_args.keys():
                            plot_args["fix_size"] = False
                        meteva.base.tool.plot_tools.scatter_sta(sta_result1, save_path=save_path1, show=show,
                                                            title=title1_list, print_max=print_max,
                                                            print_min=print_min
                                                            , add_county_line=add_county_line,
                                                             map_extend=map_extend, dpi=dpi,subplot = subplot,ncol = ncol,**plot_args)
                    else:
                        slon0 = np.min(sta_result1.loc[:, "lon"].values)
                        slat0 = np.min(sta_result1.loc[:, "lat"].values)
                        elon0 = np.max(sta_result1.loc[:, "lon"].values)
                        elat0 = np.max(sta_result1.loc[:, "lat"].values)
                        if elon0 > 180:
                            sta = sta_result1.copy()
                            sta.loc[sta_result1["lon"] > 180, "lon"] = sta_result1.loc[sta_result1["lon"] > 180, "lon"] - 360
                            slon0 = np.min(sta.loc[:, "lon"].values)
                            elon0 = np.max(sta.loc[:, "lon"].values)

                        dlon0 = (elon0 - slon0) * 0.01
                        if dlon0 > 1:
                            dlon0 = 1
                        dlat0 = (elon0 - slon0) * 0.01
                        if dlat0 > 1:
                            dlat0 = 1
                        slon = slon0 - dlon0
                        elon = elon0 + dlon0
                        slat = slat0 - dlat0
                        elat = elat0 + dlat0
                        grid1 = meteva.base.grid([slon,elon,dlon0],[slat,elat,dlat0])

                        grd = meteva.base.interp_sg_idw_delta(sta_result1,grid1,halfR=300,nearNum=8)
                        meteva.base.tool.plot_tools.contourf_2d_grid(grd, save_path=save_path1, show=show,
                                                            title=title1_list
                                                            , add_county_line=add_county_line,
                                                              dpi=dpi,subplot = subplot,ncol = ncol,**plot_args)
                if plot == "micaps":
                    meteva.base.put_stadata_to_micaps(sta_result1,layer_description=title1_list)


        if len(sta_result) == 1:
            sta_result = sta_result[0]
        if sort_by is not None:
            sta_result.sort_values(by = sort_by,axis = 0,ascending = False,inplace=True)
        result_all.append(sta_result)

    if len(result_all)==1:
        result_all = result_all[0]

    return result_all,gll1



def score_tdt(sta_ob_and_fos0,method,s = None,g = None,gll = None,group_name_list = None,
              x_y = "obtime_time",annot = 0,save_dir = None,save_path = None,show = False,
        dpi = 300,title = None,**kwargs):
    if s is not None:
        if g is not None:
            if g == "last_range" or g == "last_step":
                s["drop_last"] = False
            else:
                s["drop_last"] = True

    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s=s)
    if(len(sta_ob_and_fos1.index) == 0):
        print("there is no data to verify")
        return
    sta_ob_and_fos_list, gll1 = meteva.base.group(sta_ob_and_fos1, g=g, gll=gll)

    if "grade_list" in kwargs.keys():
        grades = kwargs["grade_list"]
        grade_names = []
        mutil_list1 = [meteva.method.ts_multi, meteva.method.bias_multi, meteva.method.ets_multi,
                       meteva.method.mr_multi, meteva.method.far_multi,
                       meteva.method.ts_grade, meteva.method.bias_grade, meteva.method.ets_grade,
                       meteva.method.mr_grade, meteva.method.far_grade]
        mutil_list2 = [meteva.method.accuracy, meteva.method.hk, meteva.method.hss]
        if method in mutil_list1:
            grade_names = ["<" + str(grades[0])]
            for i in range(len(grades) - 1):
                grade_names.append("[" + str(grades[i]) + "," + str(grades[i + 1]) + ")")
            grade_names.append(">=" + str(grades[-1]))
        elif method in mutil_list2:
            grade_names = ["0"]
        else:
            for i in range(len(grades)):
                grade_names.append(str(grades[i]))
    else:
        grade_names = ["0"]
    grade_num = len(grade_names)

    g_num = len(sta_ob_and_fos_list)
    if (g_num == 1): gll1 = [None]
    if group_name_list is None:
        group_name_list = meteva.product.program.get_group_name(gll1)

    data_names = meteva.base.get_stadata_names(sta_ob_and_fos_list[0])

    if method.__name__.find("ob_fo") >= 0:
        fo_name = data_names
        png_name = data_names[1:]
    else:
        if method.__name__.find("_uv")>=0:
            fo_name = []
            for i in range(2,len(data_names),2):
                strs = data_names[i]
                strs = strs.replace("u_","")
                fo_name.append(strs)
        elif method.__name__.find("distance")>=0:
            fo_name = []
            for i in range(2,len(data_names),2):
                strs = data_names[i]
                strs = strs.replace("lon_","")
                fo_name.append(strs)
        elif method.__name__ == "sample_count":
            fo_name = [data_names[0]]
        elif method == meteva.method.variance_mse:
            fo_name = ["variance","MSE"]
        elif method ==meteva.method.std_rmse:
            fo_name = ["STD","RMSE"]

        else:
            fo_name = data_names[1:]
        png_name = fo_name
    fo_num = len(fo_name)
    png_num = len(png_name)

    if title is not None:
        if isinstance(title, list):
            if png_num * g_num * grade_num != len(title):
                print("手动设置的title数目和要绘制的图形数目不一致")
                return


    if save_path is not None:
        if isinstance(save_path, str):
            save_path = [save_path]
        if png_num * g_num * grade_num != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return


    # get method_args and plot_args
    method_args = {}
    plot_args = {}
    method_para_list = method.__code__.co_varnames
    plot_mehod = None
    if x_y == "obtime_dtime":
        plot_mehod = meteva.base.plot_tools.mesh_time_dtime
    elif x_y == "obtime_time":
        plot_mehod = meteva.base.plot_tools.mesh_obtime_time
    elif x_y == "time_dtime":
        plot_mehod = meteva.base.plot_tools.mesh_time_dtime


    plot_para_list = []
    if plot_mehod is not None:
        plot_para_list = plot_mehod.__code__.co_varnames

    for key in kwargs.keys():
        if key in method_para_list and key not in plot_para_list:
            method_args[key] = kwargs[key]
        elif key not in method_para_list and key  in plot_para_list:
            plot_args[key] = kwargs[key]
        elif key  in method_para_list and key  in plot_para_list:
            print(method.__name__ + " and " + plot_mehod.__name__ + " have same args:" + key)
            return
        else:
            print(key + " is not args of " + method.__name__ + " or " + plot_mehod.__name__)
            return


    if "cmap" not in plot_args.keys() and "clevs" not in plot_args:
        if method == meteva.method.me:
            plot_args["cmap"] = meteva.base.cmaps.me

    result_all = []
    for k in range(g_num):
        sta_time_list,time_list = meteva.base.group(sta_ob_and_fos_list[k],g ="time")

        sta_result_list_dict = {}
        for gg in range(grade_num):
            sta_result_list_dict[gg] = []

        for st in range(len(sta_time_list)):
            sta_time = sta_time_list[st]
            result, dtime_list = score(sta_time, method, g="dtime", **method_args)

            if len(dtime_list) == 1:
                if len(result.shape) == 1:
                    if len(fo_name) > 1:
                        # 没有等级，但有多个预报成员
                        dict_result = {}
                        dict_result["dtime"] = dtime_list
                        for f in range(fo_num):
                            dict_result[fo_name[f]] = result[f]
                        sta_result = pd.DataFrame(dict_result)
                        sta_result["time"] = time_list[st]
                        sta_result_list_dict[0].append(sta_result)


                    else:
                        # 有多个等级，但只有一个预报成员
                        for gg in range(grade_num):
                            dict_result = {}
                            dict_result["dtime"] = dtime_list
                            #print(fo_name)
                            dict_result[fo_name[0]] = result[gg]
                            #print(dict_result)
                            sta_result = pd.DataFrame(dict_result)
                            sta_result["time"] = time_list[st]
                            sta_result_list_dict[gg].append(sta_result)

                else:
                    for gg in range(grade_num):
                        dict_result = {}
                        dict_result["dtime"] = dtime_list
                        for f in range(fo_num):
                            dict_result[fo_name[f]] = result[f, gg]
                        sta_result = pd.DataFrame(dict_result)
                        sta_result["time"] = time_list[st]

                        sta_result_list_dict[gg].append(sta_result)
                #print(sta_result)
            else:
                #print(dtime_list)
                #print(result)
                if len(result.shape) <=1:
                    # 没有等级，只有一个预报成员
                    if len(result.shape) ==0:
                        result = [float(result)]
                    dict_data = {"dtime":dtime_list,fo_name[0]:result}
                    #print(dict_data)
                    sta_result = pd.DataFrame(dict_data)
                    sta_result["time"] = time_list[st]
                    sta_result_list_dict[0].append(sta_result)
                elif len(result.shape) ==2:
                    if len(fo_name) > 1:
                        # 没有等级，但有多个预报成员
                        dict_result = {}
                        dict_result["dtime"] = dtime_list
                        for f in range(fo_num):
                            dict_result[fo_name[f]] = result[:,f]
                        sta_result = pd.DataFrame(dict_result)
                        sta_result["time"] = time_list[st]
                        sta_result_list_dict[0].append(sta_result)
                    else:
                        # 有多个等级，但只有一个预报成员
                        for gg in range(grade_num):
                            dict_result = {}
                            dict_result["dtime"] = dtime_list
                            #print(fo_name)
                            dict_result[fo_name[0]] = result[:,gg]
                            sta_result = pd.DataFrame(dict_result)
                            sta_result["time"] = time_list[st]
                            sta_result_list_dict[gg].append(sta_result)
                else:
                    for gg in range(grade_num):
                        dict_result = {}
                        dict_result["dtime"] = dtime_list
                        for f in range(fo_num):
                            dict_result[fo_name[f]] = result[:,f,gg]
                        sta_result = pd.DataFrame(dict_result)
                        sta_result["time"] = time_list[st]

                        sta_result_list_dict[gg].append(sta_result)

        sta_all_g_list = []

        for gg in range(grade_num):
            sta_all_g = pd.concat(sta_result_list_dict[gg])
            sta_all_g["level"] = meteva.base.IV
            sta_all_g["id"] = meteva.base.IV
            sta_all_g["lon"] = meteva.base.IV
            sta_all_g["lat"] = meteva.base.IV
            #print(sta_all_g)
            sta_all_g1 = meteva.base.sta_data(sta_all_g)
            sta_all_g_list.append(sta_all_g1)

        #print(sta_all_g_list)

        for i in range(len(sta_all_g_list)):
            sta_result1 = sta_all_g_list[i]
            sta_result1.attrs["data_source"] = "meteva." + method.__name__
            title1_list = None
            if isinstance(title, list):
                kk = k * grade_num + i
                title1_list = title[kk *png_num : (kk + 1) * png_num]
            else:
                title1_list = []
                for ii in range(png_num):
                    if title is not None:
                        title1 = meteva.product.program.get_title_from_dict(title, s, g, group_name_list[k],
                                                                            png_name[ii])
                    else:
                        title1 = meteva.product.program.get_title_from_dict("", s, g,
                                                                            group_name_list[k], png_name[ii])
                    if grade_num > 1:
                        title1 += "(grade_" + str(grade_names[i]) + ")"
                    title1_list.append(title1)


            save_path1 = None
            if save_path is None:
                if save_dir is None:
                    show = True
                else:
                    save_path1 = []
                    for i in range(len(title1_list)):
                        fileName = title1_list[i].replace("\n", "").replace(":", "")
                        save_path1.append(save_dir + "/" + fileName + ".png")
            else:
                save_path1 = save_path[k * png_num: (k + 1) * png_num]

            if method.__name__.find("ob_fo") >= 0:
                meteva.product.time_list_mesh(sta_result1,save_dir = save_dir,save_path = save_path1,show=show,dpi = dpi,title = title1_list,annot= annot,plot_error=False,**plot_args)
            else:

                #绘制图形
                #print("start plot")

                if x_y == "obtime_time":
                    meteva.base.tool.plot_tools.mesh_obtime_time(sta_result1,save_dir=save_dir,save_path=save_path1,show = show,dpi = dpi,title = title1_list,annot = annot,**plot_args)
                elif x_y == "obtime_dtime":
                    meteva.base.tool.plot_tools.mesh_obtime_dtime(sta_result1, save_dir=save_dir,save_path=save_path1,show = show,dpi = dpi, title=title1_list,annot = annot,**plot_args)
                elif x_y == "time_dtime":
                    #print(sta_result1)
                    meteva.base.tool.plot_tools.mesh_time_dtime(sta_result1, save_dir=save_dir,save_path=save_path1,show = show,dpi = dpi, title=title1_list,annot = annot,**plot_args)
                else:
                    print("目前绘图样式参数 x_y仅支持 obtime_time, obtime_dtime, time_dtime三种形式")
        if len(sta_all_g_list) == 1:
            sta_all_g_list = sta_all_g_list[0]


        result_all.append(sta_all_g_list)

    if len(result_all) == 1:
        result_all = result_all[0]
    return result_all, gll1


def score_obhour_dtime(sta_ob_and_fos0,method,s = None,g = None,gll = None,group_name_list = None,
              x_y = "obtime_time",annot = 0,save_dir = None,save_path = None,show = False,
        dpi = 300,title = None,**kwargs):
    pass


# -*- coding: utf-8 -*-
"""
Created on Tue Mar  2 11:31:55 2021

@author: 1
"""
from pynverse import inversefunc  # 求逆函数的包
import numpy as np
import pandas as pd
import math


def vxstats(x, xhat, which_stats=None, subset=None):
    if which_stats is None:
        which_stats = ["bias", "ts", "ets", "pod", "far", "f", "hk", "bcts", "bcets"]
    out = {}
    xdim = xhat.shape
    # 将缺省值填为0
    if type(x) == pd.DataFrame:
        x = x.fillna(0)
    if type(xhat) == pd.DataFrame:
        xhat = xhat.fillna(0)

    if "mse" in which_stats:
        if subset is None:
            Nxy = np.sum(np.sum((xhat != np.nan) & (x != np.nan), axis=0))
            out['mse'] = np.sum(np.sum((xhat - x) ** 2, axis=0)) / Nxy
        else:
            out['mse'] = np.mean((xhat[subset] - x[subset]) ** 2)
    elementExist = []
    for i, value in enumerate(which_stats):
        elementExist.append(value in ['bias', 'ts', 'ets', 'pod', 'far', 'f', "hk", "bcts", "bcets"])
    if any(elementExist):
        if xhat is not bool:
            xhat = xhat != 0
        if x is not bool:
            x = x != 0
        if subset is None:
            hits = np.sum(np.sum(xhat & x))
            miss = np.sum(np.sum((~xhat) & x))
            fa = np.sum(np.sum(xhat & (~x)))
            elementExist2 = []
            for i, value in enumerate(["ets", "f", "hk"]):
                elementExist2.append(value in which_stats)
            if any(elementExist2):
                cn = np.sum(np.sum((~xhat) & (~x)))
        else:
            # hits <- sum(c(Xhat)[subset] & c(X)[subset], na.rm = TRUE)
            hits = np.sum(np.sum(xhat[subset] & x[subset]))
            # miss <- sum(!c(Xhat)[subset] & c(X)[subset], na.rm = TRUE)
            miss = np.sum(np.sum((~xhat[subset]) & x[subset]))
            # fa <- sum(c(Xhat)[subset] & !c(X)[subset], na.rm = TRUE)
            fa = np.sum(np.sum(xhat[subset] & (~x[subset])))
            elementExist2 = []
            for i, value in enumerate(["ets", "f", "hk"]):
                elementExist2.append(value in which_stats)
            if any(elementExist2):
                cn = np.sum(np.sum((~xhat[subset]) & (~x[subset])))
        if "bias" in which_stats:
            if (hits + fa == 0) and (hits + miss == 0):
                out['bias'] = 1
            else:
                out['bias'] = (hits + fa) / (hits + miss)
        if "ts" in which_stats:
            if hits == 0:
                out['ts'] = 0
            else:
                out['ts'] = hits / (hits + miss + fa)
        if "ets" in which_stats:
            if ((hits + miss == 0) or (hits + fa == 0)):
                hits_random = 0
            else:
                hits_random = float(hits + miss) * float(hits + fa) / float(hits + miss + fa + cn)
            if (hits + miss + fa == 0):
                out['ets'] = 0
            else:
                out['ets'] = (hits - hits_random) / (hits + miss + fa - hits_random)
        elementExist3 = []
        for i, value in enumerate(["pod", "hk"]):
            elementExist3.append(value in which_stats)
        if any(elementExist3):
            if (hits + miss == 0):
                pod = 0
            else:
                pod = hits / (hits + miss)
            if "pod" in which_stats:
                out['pod'] = pod
        if "far" in which_stats:
            if (hits + fa == 0):
                out['far'] = 0
            else:
                out['far'] = fa / (hits + fa)
        elementExist4 = []
        for i, value in enumerate(["f", "hk"]):
            elementExist4.append(value in which_stats)
        if any(elementExist4):
            if (cn + fa == 0):
                f = 0
                out['f'] = f
            else:
                f = fa / (cn + fa)
                out['f'] = fa / (cn + fa)
            if "hk" in which_stats:
                f = pod - f
                out['hk'] = f
        elementExist5 = []
        for i, value in enumerate(["bcts", "bcets"]):
            elementExist5.append(value in which_stats)
        if any(elementExist5):
            nF = hits + fa
            nO = hits + miss
            lf = math.log(nO / miss)
            LambertW = (lambda y: y * math.exp(y))
            # Ha = nO - (fa/lf) * LambertW((nO/fa) * lf)
            Ha = nO - (fa / lf) * inversefunc(LambertW, y_values=((nO / fa) * lf))
            if "bcts" in which_stats:
                out['bcts'] = Ha / (2 * nO - hits)
            if "bcets" in which_stats:
                out['bcets'] = (Ha - (nO ** 2) / (hits + miss + fa +
                                                  cn)) / (2 * nO - Ha - (nO ** 2) / (hits + miss + fa + cn))
    out['class'] = "vxstats"
    return out
